from malevich_space.ops import SpaceOps

from .space import resolve_setup


def get_auto_ops() -> SpaceOps:
    from malevich.manifest import ManifestManager

    return SpaceOps(resolve_setup(
        ManifestManager().query('space', resolve_secrets=True)
    ))



from typing import Any


class InterpretationError(Exception):
    def __init__(
        self, message: str, interpreter: 'Interpreter' = None, state: Any = None  # noqa: ANN401
    ) -> None:
        super().__init__(message)
        if interpreter:
            self.interpreter = interpreter
            self.state = state or interpreter.state


class NoPipelineFoundError(Exception): ...
class NoTaskToConnectError(Exception): ...
class ForbiddenSyntaxError(SyntaxError): ...

class NewVariableInBranchError(SyntaxError):
    message = (
        "Malevich @flow definition does not support branches "
        "that introduces a new variable. Declare the variable "
        "in the upper-level of the function "
        "\n@flow"
        "\ndef {function_name}(...):"
        "\n\t{name} = None"
        "\n\t# ..."
        "\n"
        "\n or consider using existing variables in the branch. "
        "Variable: {name}"
    )
    
    def finalize(
        self,
        filename: str,
        function_name: str | None = None,
        var_name: str | None = None
    ):
        self.args = (
            self.message.format(
                function_name=function_name,
                name=var_name or self.var_name
            ),
            (filename, *self.details[1:])
        )
        super().__init__(*self.args)
        return self
    
    def __init__(
        self,
        var_name: str,
        function_name: str | None,
        details,
    ):
        self.var_name = var_name
        self.details = details
        if function_name:
            self.args[0] = self.message.format(
                function_name=function_name,
                name=var_name
            )



from sqlalchemy import Column, String, Integer, Sequence, DateTime, ForeignKey
from .base import Base


class Artifact(Base):
    __tablename__ = 'artifacts'
    
    id = Column(Integer, Sequence('artifact_seq'), primary_key=True)
    payload = Column(String)
    captured_at = Column(DateTime)
    
    session_id = Column(Integer, ForeignKey('artifacts_sessions.id'))
    
class Session(Base):
    __tablename__ = 'artifacts_sessions'
    
    id = Column(Integer, Sequence('session_seq'), primary_key=True)
    start = Column(DateTime, nullable=False)
    end = Column(DateTime)
    python_version = Column(String)
    malevich_version = Column(String)
    
    master_id = Column(Integer, ForeignKey('artifacts_master_sessions.id'))

class MasterSession(Base):
    __tablename__ = 'artifacts_master_sessions'
    
    id = Column(Integer, Sequence('master_session_seq'), primary_key=True)    

   


from typing import Iterable, Union

from malevich._autoflow.tracer import traced
from malevich._autoflow.tree import ExecutionTree
from malevich.models import ArgumentLink, BaseNode

Nodes = Iterable[BaseNode]
TracedNode = traced[BaseNode]
TracedNodes = Iterable[traced[BaseNode]]
FlowOutput = Union[TracedNode, TracedNodes, None]
FlowTree = ExecutionTree[traced[BaseNode], ArgumentLink]



from typing import Callable, TypeVar

ActionResult = TypeVar('ActionResult')
AttemptResult = TypeVar('AttemptResult')
FinalResult = TypeVar('FinalResult')

def try_cascade(
    action: Callable[[], ActionResult],
    *attempts: Callable[[Exception], AttemptResult],
    exceptions: tuple[Exception | KeyboardInterrupt] | None = None
) -> tuple[ActionResult | AttemptResult, Exception | KeyboardInterrupt | None]:
    if not exceptions:
        exceptions = (Exception,)
    try:
        result = action()
    except exceptions as e:
        for attempt in attempts:
            try:
                result = attempt(e)
                if result:
                    return action(), None
            except exceptions as e:
                pass
            else:
                return result, None
        return result, e
    else:
        return result, None



from typing import (
    Callable,
    Generic,
    ParamSpec,
    Type,
    TypeVar,
)

from pydantic import BaseModel

from malevich._autoflow.function import autotrace, sinktrace
from malevich.constants import reserved_config_fields
from malevich.models import ConfigArgument

FnArgs = ParamSpec("FnArgs")
FnReturn = TypeVar("FnReturn")
Config = TypeVar("Config", bound=BaseModel)
ProcConfig = TypeVar("ProcConfig", bound=BaseModel)

ProcFunArgs = ParamSpec("ProcFunArgs")
ProcFunReturn = TypeVar("ProcFunReturn")

class ConfigStruct(dict):
    def __new__(cls, **kwargs):
        return dict(**kwargs)

    def __init__(self, **kwargs) -> None:
        """Creates a configuration from keyword arguments"""


class ProcessorFunction(Generic[Config, ProcFunArgs, ProcFunReturn]):
    def __init__(
        self,
        fn: Callable[FnArgs, FnReturn],
        use_sinktrace: bool = False,
        config_model: Type[Config] = BaseModel,
    ) -> None:
        self.base_fn = fn
        if use_sinktrace:
            self.__fn = sinktrace(fn)
        else:
            self.__fn = autotrace(fn)
        self.__config_model = config_model
        self.__call__ = self._fn_call
        self.__required_fields = [
            key
            for key, value in self.__fn.__annotations__.items()
            if hasattr(value, '__metadata__')
            and (metadata := getattr(value, '__metadata__'))
            and isinstance(metadata[0], ConfigArgument)
            and metadata[0].required
        ]

    def _fn_call(self, *args: FnArgs.args, **kwargs: FnArgs.kwargs) -> FnReturn:
        # FIXME: literal 'config' is overused. Should be replaced with constant.
        if kwargs.get('config', None) is None:
            kwargs['config'] = {}

        if isinstance(kwargs['config'], BaseModel):
            assert type(kwargs['config']) == self.__config_model, (  # noqa: E721
                f"You have set config={kwargs['config']}, "
                f"but it should be of type {self.__config_model}."
            )

            kwargs['config'] = kwargs['config'].model_dump()
        else:
            assert isinstance(kwargs, dict)
        reserved_keys = {
            x[0] for x in reserved_config_fields
        }
        extra_fields = {**kwargs}
        extra_fields.pop('config')
        for reserved_keys, _ in reserved_config_fields:
            extra_fields.pop(reserved_keys, None)

        kwargs['config'] = {
            **kwargs['config'],
            **extra_fields,
        }

        kwargs = {
            'config': kwargs['config'],
            **{
                k: v for k, v in kwargs.items()
                if k in reserved_keys
            }
        }

        if (diff_ := set.difference(set(self.__required_fields), kwargs['config'].keys())) != set():  # noqa: E501
            fields_ = ', '.join([f"`{x}`" for x in diff_])
            raise Exception(
                f"Missing required fields {fields_}"
                f" in configuration of the processor `{self.__fn.__name__}`"
            )

        return self.__fn(*args, **kwargs)

    __call__: Callable[ProcFunArgs, ProcFunReturn] = _fn_call

    @property
    def __doc__(self) -> str:
        return self.base_fn.__doc__

    @property
    def config(self) -> Type[Config] | Type[ConfigStruct]:
        return self.__config_model or ConfigStruct


def proc(
    use_sinktrace: bool = False,
    config_model: Type[ProcConfig] = BaseModel,
) -> Callable[
    [Callable[FnArgs, FnReturn]],
    Callable[FnArgs, FnReturn] | ProcessorFunction[ProcConfig, FnArgs, FnReturn]
]:
    def decorator(
        fn: Callable[FnArgs, FnReturn]
    ) -> ProcessorFunction[ProcConfig, FnArgs, FnReturn]:
        return ProcessorFunction[ProcConfig, FnArgs, FnReturn](
            fn, use_sinktrace, config_model
        )

    return decorator



from malevich._autoflow.tracer import traced
from malevich.models.nodes.operation import OperationNode


class AddPositiveCondition:
    def __init__(self, uid) -> None:
        self.uid = uid

    def __call__(self, x: traced[OperationNode], from_node: bool | None = None) -> None:
        if from_node is not None and from_node:
            return
        if not isinstance(x.owner, OperationNode):
            return
        if self.uid not in x.owner.should_be_true:
            x.owner.should_be_true = [*x.owner.should_be_true, self.uid]


class AddNegativeCondition:
    def __init__(self, uid) -> None:
        self.uid = uid

    def __call__(self, x: traced[OperationNode], from_node: bool | None = None) -> None:
        if from_node is not None and from_node:
            return
        if not isinstance(x.owner, OperationNode):
            return
        if self.uid not in x.owner.should_be_false:
            x.owner.should_be_false = [*x.owner.should_be_false, self.uid]




"""
Provides functionality for creating a package with the given name and contents.
"""


import importlib
import os
from hashlib import sha256
from pathlib import Path

import malevich

from ..models.dependency import Dependency


def create_package_stub(
    package_name: str,
    metascript: str,
    dependency: Dependency
) -> tuple[bool, str]:
    """
    Create a package with the given name and contents
    Args:
        package (str): The name of the package to create.
        metascript (str): The contents of the package.

    Returns:
        A tuple containing a boolean indicating whether the package was
        created successfully, and the checksum of the package.
    """
    # Calculate the checksum of the package contents
    checksum = sha256(metascript.encode()).hexdigest()

    # Create the package directory
    # The package will be located at <python>/malevich/<package>
    package_path = Path(malevich.__file__).parent / package_name
    # Create the package directory if it doesn't exist
    os.makedirs(package_path, exist_ok=True)

    # Create the package __init__.py file
    with open(package_path / "__init__.py", "w") as f:
        # Write the metascript to the file
        f.write(metascript)
        # Write the checksum to the file
        f.write(f"\n__Metascript_checksum__ = '{checksum}'\n")
        f.write(f"\n__Dependency_checksum__ = '{dependency.checksum()}'\n")

    # Post-hoc check to make sure the package was created successfully
    try:
        # Import the package
        __pkg = importlib.import_module(f"malevich.{package_name}")
        # Assert that the checksum of the package is the same as the checksum
        assert getattr(__pkg, "__Metascript_checksum__") == checksum, \
            "Package checksum does not match metascript checksum. "
    except (ImportError, AssertionError):
        return False, checksum
    return True, checksum



import base64
import tempfile
from typing import Any

import pandas as pd
from malevich_space.schema.schema import SchemaMetadata

from malevich.models import Collection

from .base import BaseNode


class CollectionNode(BaseNode):
    collection: Collection
    scheme: str | SchemaMetadata | None = None

    def get_senstivite_fields(self) -> dict[str, Any]:
        # Save pd.DataFrame as binary data and encode
        # it with base64 to save as string
        data = self.collection.collection_data
        with tempfile.TemporaryFile() as f:
            data.to_pickle(f)
            f.seek(0)
            encoded = base64.b64encode(f.read()).decode()
        if data is None:
            return {}
        return {"collection_data": encoded}

    def set_sensitive_fields(self, values: dict[str, str]) -> None:
        collection_data = values.get("collection_data")
        if collection_data is None:
            return
        decoded = base64.b64decode(collection_data)
        with tempfile.TemporaryFile() as f:
            f.write(decoded)
            f.seek(0)
            data = pd.read_pickle(f)
        self.collection.collection_data = data

    def __hash__(self) -> int:
        return super().__hash__()



import inspect
import json
import warnings
from functools import wraps
from inspect import signature
from typing import Any, Callable, Optional, ParamSpec, TypeVar, overload

import pandas as pd

from malevich._autoflow import Flow, traced
from malevich._utility import Registry, generate_empty_df_from_schema
from malevich.models import (
    ArgumentLink,
    FlowFunction,
    PromisedTask,
    TreeNode,
)
from malevich.types import TracedNode, TracedNodes

from .._ast import boot_flow
from ..models.nodes.morph import MorphNode
from .collection import collection

T = TypeVar("T")
FlowDecoratorReturn = PromisedTask | TracedNode | TracedNodes
Args = ParamSpec("Args")
reg = Registry()

class NoArgumentAnnotation(Warning): ...
class NotTracedArgumentError(Exception): ...
class SubflowArgumentIsDataFrameError(Exception): ...

@overload
def flow(
    fn = None,
    /,
) -> FlowFunction[Args, FlowDecoratorReturn]:
    pass

@overload
def flow(
    *,
    reverse_id: Optional[str] = None,
    name: Optional[str] = None,
    description: Optional[str] = None,
    dfs_are_collections: Optional[bool] = None,
) -> Callable[[Callable[Args, T]], FlowFunction[Args, FlowDecoratorReturn]]:
    pass

def flow(
    fn = None,
    *,
    reverse_id: Optional[str] = None,
    name: Optional[str] = None,
    description: Optional[str] = None,
    **kwargs: Any,
) -> Callable[[Callable[Args, T]], FlowFunction[Args, FlowDecoratorReturn]]:
    """Converts a function into a flow

    The function is converted into :class:`FlowFunction <malevich.models.flow_function.FlowFunction>` object that can be
    called to produce a task or serve as a subflow.

    When the function is called from within another :code:`@flow()` decorated
    function, it returns traced objects to be used as arguments for other
    flow components. When the function is called from the main context, it
    returns a task that can be interpreted by any of available interpreters.

    Args:
        reverse_id (str, optional): Reverse ID of the flow component. Defaults to the name of the function.
        name (str, optional): Name of the flow. Defaults to None.
        description (str, optional): Description of the flow. Defaults to None.
        dfs_are_collections (bool, optional): Whether to treat pandas.DataFrame as a collection. Defaults to False.
        **kwargs (Any):
            Additional arguments to be passed to the flow component.
            See :class:`malevich_space.schema.ComponentSchema` for details.

    Returns:
        Callable[[Callable[Args, T]], Callable[Args, T]]: Decorator for the function.
    """   # noqa: E501

    globals_ = inspect.currentframe().f_back.f_globals
    local_vars = inspect.currentframe().f_back.f_locals

    def wrapper(function: Callable[Args, T]) -> FlowFunction[Args, FlowDecoratorReturn]:
        nonlocal reverse_id, name, description
        reverse_id = reverse_id or function.__name__
        function_name = name or function.__name__.replace("_", " ").title()
        description = description or function.__doc__

        if description is None:
            description = f"Meta flow: {function_name}"

        annotations = {}
        sign = signature(function)

        # Check whether the function has variable positional arguments
        # or variable keyword arguments
        if any(
            p.kind in (p.VAR_POSITIONAL, p.VAR_KEYWORD)
            for p in sign.parameters.values()
        ):
            raise TypeError(
                "Functions with variable positional or keyword arguments "
                "cannot be converted into flows"
            )

        # Check whether the function has positional-only arguments
        if any(p.kind == p.POSITIONAL_ONLY for p in sign.parameters.values()):
            raise TypeError(
                "Functions with positional-only arguments "
                "cannot be converted into flows"
            )

        # Check whether the function has annotations
        for p in sign.parameters.values():
            if p.annotation is p.empty:
                annotations[p.name] = collection[p.name, None]
                warnings.warn(
                    f"Argument '{p.name}' has no annotation. "
                    f"Assuming it is a collection with name '{p.name}'. "
                    "It is recommended to provide an explicit annotation.",
                    NoArgumentAnnotation,
                )
            else:
                annotations[p.name] = p.annotation

        pos_arg_names = [
            p.name for p in sign.parameters.values() if p.default is p.empty
        ]

        @wraps(function)
        def fn(*args: Args.args, __component, **kwargs: Args.kwargs) -> FlowDecoratorReturn:
            is_subflow = Flow.isinflow()
            args = list(args)
            if is_subflow:
                outer_tracer = traced()

            with Flow():
                named_args = {
                    **{name: value for name, value in zip(pos_arg_names, args)},
                    **kwargs
                }

                traced_args: dict[str, traced] = {}
                traced_shadows: dict[str, traced] = {}
                non_traced_args: dict[str, Any] = {}

                for name, value in named_args.items():
                    if is_subflow and not isinstance(value, traced):
                      raise NotTracedArgumentError(
                          f"You have called a flow function {function.__name__} "
                          "from within another flow function, but the argument "
                          f"{name} is not a valid object. You can call the function "
                          "only with arguments that are results of Malevich operations "
                          "such as other flow functions, collections, or processors."
                      )
                for (name, value), param in zip(
                    named_args.items(), sign.parameters.values()
                ):
                    ann = annotations[name]
                    if hasattr(ann, "__malevich_collection_name__") and (
                        isinstance(value, pd.DataFrame)
                    ):
                        collection_name = ann.__malevich_collection_name__
                        collection_scheme = ann.__malevich_collection_scheme__  # noqa: E501
                    else:
                        raise TypeError(
                            f"Argument '{name}' has invalid annotation. "
                            "It should be a `collection[name]` or nothing."
                        )

                    if isinstance(value, pd.DataFrame):
                        if is_subflow:
                            raise SubflowArgumentIsDataFrameError(
                                "You have called a flow function "
                                f"{function.__name__} from within another flow "
                                f"function, but the argument {name} is a DataFrame. "
                                "You can call the function only with arguments "
                                "that are results of Malevich operations such as "
                                "other flow functions, collections, or processors. "
                                "If you want to pass a DataFrame to the function, "
                                "you should call it from the main context."
                            )
                        else:
                            traced_args[name] = (
                                collection(
                                    df=value,
                                    name=collection_name,
                                    alias=name
                                )
                            )

                            # Substitute with collection for args as well
                            for i in range(len(args)):
                                if args[i] is value:
                                    args[i] = traced_args[name]

                            for k in kwargs:
                                if kwargs[k] is value:
                                    kwargs[k] = traced_args[name]


                    elif isinstance(value, traced):
                        df_ = generate_empty_df_from_schema(
                            collection_scheme
                        )

                        traced_args[name] = collection(
                            collection_name,
                            alias=name,
                            df=df_,
                            scheme=json.dumps(collection_scheme)
                        )

                        traced_shadows[name] = value
                    else:
                        non_traced_args[name] = value

                if is_subflow:
                    traced_pos_args = [
                        traced_args[name] for name in pos_arg_names
                    ]
                    non_traced_pos_args = [
                        non_traced_args[name] for name in pos_arg_names
                    ]
                    traced_kwargs = {
                        name: traced_args[name] for name in kwargs
                        if name not in pos_arg_names
                    }
                    non_traced_kwargs = {
                        name: non_traced_args[name] for name in kwargs
                        if name not in pos_arg_names
                    }

                    __results = boot_flow(
                        function,
                        globals_,
                        local_vars,
                        *[*non_traced_pos_args, *traced_pos_args],
                        **{**non_traced_kwargs, **traced_kwargs}
                    )
                else:
                    __results = boot_flow(
                        function,
                        globals_,
                        local_vars,
                        *args,
                        **kwargs
                    )

                sub_tree = Flow.flow_ref()

                sub_tree.cast_link_types(
                    lambda x: ArgumentLink(
                        **x.model_dump()
                    ) if not isinstance(x, ArgumentLink) else x
                )

                t_node = TreeNode(
                    tree=sub_tree,
                    reverse_id=reverse_id,
                    name=function_name,
                    description=description,
                    results=__results or []
                )

            # The magic starts here
            # ---------------------

            # Case 1: The function is called from the flow
            if is_subflow:
                # Attaching the outer tracer to the node
                outer_tracer.claim(t_node)

                for index, (name, value) in enumerate(named_args.items()):
                    if isinstance(value, traced):
                        # Obtain the bridge nodes: one that connects the argument
                        bridges = sub_tree.edges_from(traced_args[name])
                        bridge_nodes = [(b[2], b[1]) for b in bridges]
                        # Now, bridges are connected
                        _a = ArgumentLink(
                            index=index,
                            name=name,
                            is_compressed_edge=True,
                            compressed_edges=bridge_nodes,
                            shadow_collection=traced_args[name]
                        )
                        traced_shadows[name]._autoflow.calledby(outer_tracer, _a)

                # Check whether len of each item in __results is the same
                for r in __results:
                    if len(r) != len(__results[0]):
                        raise ValueError(
                            "Subflow function returned different number of results "
                            "in return statements. All return statements should return "
                            "the same number of results."
                        )

                morphed_results = []
                len_ = len(__results[0][1]) if isinstance(__results[0][1], tuple) else 1
                for i in range(len_):
                    morph = MorphNode()
                    for morphome in __results:
                        morph.members.append((morphome[0], morphome[1][i].owner))
                    morph.correct_self()
                    morphed_results.append(morph)

                outputs = [
                    traced(
                        TreeNode(
                            **t_node.model_dump(exclude=["underlying_node"]),
                            underlying_node=o,
                        )
                    )
                    for o in morphed_results
                ]

                assert all([o.owner.results is not None for o in outputs])
                return outputs if len(outputs) > 1 else outputs[0]
            else:
                # Save t_node.tree
                return PromisedTask(
                    results=__results, tree=t_node, component=__component
                )

        return FlowFunction(fn, reverse_id, function_name, description, **kwargs)

    return wrapper(fn) if fn else wrapper



from typing import Generic, TypeVar

from ..._utility.core_logging import IgnoreCoreLogs

CreateFn = TypeVar('CreateFn', bound=callable)
DeleteFn = TypeVar('DeleteFn', bound=callable)
UpdateFn = TypeVar('UpdateFn', bound=callable)
GetFn = TypeVar('GetFn', bound=callable)
ListFn = TypeVar('ListFn', bound=callable)

T = TypeVar('T')
def wrap_fn(
    fn: T,
    ref: 'BaseRef',
    create: bool = False,
    delete: bool = False,
    update: bool = False,
    get: bool = False,
    list: bool = False,
) -> T:

    def wrapped(*args, **kwargs):
        if create:
            out = fn(*args, **kwargs)
            ref._created = True
            ref._deleted = False
            return out
        if delete:
            out = fn(*args, **kwargs)
            ref._deleted = True
            ref._created = False
            return out
        return fn(*args, **kwargs)
    return wrapped


def create_if_not_exists(
    ref: 'BaseRef',
    create: CreateFn,
) -> CreateFn:
    def wrapped(*args, **kwargs):
        with IgnoreCoreLogs():
            if (_get := ref._try_get()):
                return _get
        return create(*args, **kwargs)
    return wrapped

def update_or_create(
    ref: 'BaseRef',
    create: CreateFn,
    update: UpdateFn,
) -> UpdateFn:
    def wrapped(*args, **kwargs):
        if ref.update is not None:
            try:
                return update(*args, **kwargs)
            except Exception as e:
                if ref.create is not None:
                    return create(*args, **kwargs)
                else:
                    raise ValueError(
                        f'Failed to update {ref} and '
                        'no create function is available'
                    )
        if ref.create is not None:
            return create(*args, **kwargs)
        else:
            raise ValueError(
                f'No create function is available for {ref}'
            )
    return wrapped

class BaseRef(Generic[CreateFn, DeleteFn, UpdateFn, GetFn, ListFn]):
    def __init__(
        self,
        name: str,
        create: CreateFn | None = None,
        delete: DeleteFn | None = None,
        update: UpdateFn | None = None,
        get: GetFn | None = None,
        list: ListFn | None = None,
    ) -> None:
        self.name = name
        self.create = wrap_fn(create, self, create=True)
        self.delete = wrap_fn(delete, self, delete=True)
        self.update = update
        self.get = get
        self.list = list

        self._created = None
        self._deleted = None

    def _try_get(self):
        if self.get is None:
            return None
        try:
            return self.get() is not None
        except Exception as e:
            return False


    @property
    def created(self) -> None | bool:
        return self._created or self._try_get()


    @property
    def deleted(self) -> None | bool:
        return self._deleted or not self._try_get()

    @property
    def create_if_not_exists(self) -> CreateFn | None:
        return create_if_not_exists(self, self.create)

    @property
    def update_or_create(self) -> UpdateFn | None:
        return update_or_create(self, self.create, self.update)

    def __str__(self) -> str:
        return self.name




from .base import BaseNode
from .collection import CollectionNode
from .operation import OperationNode
from .tree import TreeNode
from .asset import AssetNode
from .document import DocumentNode
from .morph import MorphNode



from sqlalchemy import create_engine
from sqlalchemy.orm.session import Session

from malevich.path import Paths
from ..schema import Base


def get_db() -> Session:
    db_ref = globals().get('__Malevich_db__', None)
    
    if not db_ref:
        import duckdb
        # HACK: Create empty duckdb db
        duckdb.connect(Paths.db()).commit().close()
        engine = create_engine("duckdb:///" + Paths.db())
        Base.metadata.create_all(engine)
        db_ref = engine
        globals()['__Malevich_db__'] = engine

    return Session(bind=db_ref)
        



import re
from collections import defaultdict
from typing import overload

from malevich_space.ops.space import SpaceOps
from malevich_space.schema import (
    LoadedFlowSchema,
    LoadedInFlowComponentSchema,
    SpaceSetup,
)

from malevich._autoflow.tree import ExecutionTree
from malevich._utility import resolve_setup
from malevich.manifest import manf


class SpaceFlowExporter:
    @overload
    def __init__(self, reverse_id: str, setup: SpaceSetup | None) -> None:
        pass

    @overload
    def __init__(self, flow: LoadedFlowSchema) -> None:
        pass

    def __init__(
        self,
        reverse_id: str | None = None,
        setup: SpaceSetup | None = None,
        flow: LoadedFlowSchema | None = None
    ):
        assert reverse_id is not None or flow is not None, (
            'Requires `flow` or `reverse_id` to be not None'
        )
        self.reverse_id = reverse_id
        self._alias_factory = defaultdict(lambda: 1)
        self._alias_memory = set()

        if flow is not None:
            self.flow = flow
            return

        if setup is None:
            setup = resolve_setup(manf.query("space", resolve_secrets=True))
        ops = SpaceOps(setup)

        component_ = ops.get_parsed_component_by_reverse_id(reverse_id=reverse_id)

        if component_ is None or component_.flow is None:
            raise ValueError(f"{reverse_id} does not exists or not a flow")

        self.flow = component_.flow

    def _make_alias(self, component: LoadedInFlowComponentSchema):
        x = f'{component.reverse_id}_{self._alias_factory[component.reverse_id]}'
        while x in self._alias_memory:
            self._alias_factory[component.reverse_id] += 1
            x = f'{component.reverse_id}_{self._alias_factory[component.reverse_id]}'
        self._alias_memory.add(x)
        return x

    def _varname(self, component: LoadedInFlowComponentSchema)-> str:
        # TODO: Make with LLM
        return re.sub(
            r'[\W\s]+',
            '_',
            component.alias or component.reverse_id
        ).lower()

    def _decl(self, varname: str, fn: str, *args, **kwargs):
        kwargs_ = [f'{k}={v!r}' for k, v in kwargs.items()]
        args_ = ', '.join([*map(repr, args), *kwargs_])

        decl = f'{varname} = {fn}({args_})'
        return decl

    def get_meta_pipeline(
        self,
        include_def: bool = True,
        include_decorator: bool = True,
        include_return: bool = True,
        reverse_id: str | None = None,
    ):
        assert reverse_id or self.reverse_id, (
            '`reverse_id` is required if not set in the constructor'
        )
        reverse_id = reverse_id or self.reverse_id
        apps = set([x.reverse_id for x in self.flow.components if x.app is not None])

        body = []
        for component in self.flow.components:
            if not component.alias:
                component.alias = self._make_alias(component)
            else:
                self._alias_memory.add(component.alias)

        varnames = {
            x.alias: self._varname(x)
            for x in self.flow.components
        }
        uid2alias = {}
        for component in self.flow.components:
            uid2alias[component.uid] = component.alias

        collections = [
            x
            for x in self.flow.components
            if x.collection is not None
        ]

        body.extend([
            (node.uid, self._decl(
                varnames[node.alias],
                'collection',
                node.reverse_id,
                alias=node.alias,
            ))
            for node in collections
        ])

        operations = [
            (x, *x.prev)
            for x in self.flow.components
            if x.app is not None
        ]

        config_repr = {
            x.uid: x.active_cfg.cfg_json if x.active_cfg is not None else {}
            for x in self.flow.components
        }

        body.extend([
            (
                node.uid,
                self._decl(
                    varnames.get(node.alias),
                    node.app.active_op[0].name,
                    *[varnames.get(uid2alias.get(x.uid)) for x in y],
                    **config_repr.get(node.uid, {}),
                )
            )
            for node, *y in operations
        ])

        instructions = dict(body)
        tree = ExecutionTree()
        for node in self.flow.components:
            for index, prev in enumerate(node.prev):
                tree.put_edge(prev.uid, node.uid, index)

        ordered_instructions = [
            instructions[x]
            for x in tree.topsort()
        ]

        if include_return:
            ordered_instructions.append(
                'return ' + ', '.join([
                    varnames[uid2alias[x]] for x in tree.leaves()
                ])
            )

        imports_ = (
            'from malevich import flow, collection, table\n'
            + '\n'.join(
                f'from malevich.{x} import *'
                for x in apps
            )
        )


        imports_ = (
            'from malevich import flow, collection, table\n'
            + '\n'.join(
                f'from malevich.{x} import *'
                for x in apps
            )
        )

        body_ = '\n\t'.join(ordered_instructions)
        reverse_id_ = re.sub(r"[\W\s]+", "_", reverse_id).lower()
        def_ = f'def {reverse_id_}():'
        if include_decorator:
            def_ = f'@flow\n{def_}'
        if include_def:
            return f'{def_}\n\t{body_}', imports_
        return body_, imports_



from typing import Annotated

from pydantic import AfterValidator


def _pystr_validate(_str: str) -> str:
    if not isinstance(_str, str):
        raise ValueError('Invalid Python variable name. Expected a string')
    if not _str.isidentifier():
        raise ValueError(
            'Invalid Python variable name. Must be a valid Python identifier'
        )
    return _str

PythonString = Annotated[str, AfterValidator(_pystr_validate)]



from pathlib import Path
import re
from typing import Annotated, Optional

import rich
import typer
from malevich_space.schema import LoadedComponentSchema
from rich.progress import Progress, SpinnerColumn, TextColumn

from malevich._deploy import Space
from malevich.constants import (
    DEFAULT_CORE_HOST,
    IMAGE_BASE,
)
from malevich.install import FlowInstaller, ImageInstaller, SpaceInstaller
from malevich.install.local import LocalInstaller
from malevich.manifest import ManifestManager
from malevich.models.dependency import Integration

from ..help import use as help

use = typer.Typer(help=help['--help'], rich_markup_mode="rich")


def _install_from_image(
    package_name: str,
    image_ref: Optional[str] = None,
    image_auth_user: Optional[str] = None,
    image_auth_password: Optional[str] = None,
    core_host: str = DEFAULT_CORE_HOST,
    core_user: Optional[str] = None,
    core_token: Optional[str] = None,
    progress: Progress = None,
) -> None:
    """UI-friendly wrapper for image installer

    Install package from image registry and renders CLI interface
    for the process of installation.

    Use `~ImageInstaller.install` method to directly install package

    Args:
        package_name (str): Name of the package
        image_ref (Optional[str], optional): Image reference in format
            <registry>/<image>:<tag>. Defaults to None.
        image_auth_user (Optional[str], optional): Image registry auth user.
            Defaults to None.
        image_auth_password (Optional[str], optional): Image registry auth password.
            Defaults to None.
        core_host (str, optional): Malevich Core hostname.
            Defaults to DEFAULT_CORE_HOST.
        core_user (Optional[str], optional): Malevich Core user. Defaults to None.
        core_token (Optional[str], optional): Malevich Core token. Defaults to None.
        progress (Progress, optional): Rich progress bar. Defaults to None.

    Raises:
        Exception: If package with the same name already exists in manifest
            with different installer
    """
    if image_ref is None:
        return _install_from_image(
            package_name=package_name,
            image_ref=IMAGE_BASE.format(app=package_name),
            image_auth_user=image_auth_user,
            image_auth_password=image_auth_password,
            core_host=core_host or DEFAULT_CORE_HOST,
            core_user=core_user,
            core_token=core_token,
            progress=Progress(
                SpinnerColumn(), TextColumn("{task.description}")),
        )
    installer = ImageInstaller()
    if progress:
        task_ = progress.add_task(
            f"Attempting to install [b blue]{package_name}[/b blue] with"
            " [i yellow]image[/i yellow] installer",
            total=1,
        )

    try:
        manifest_entry = installer.install(
            package_name=package_name,
            image_ref=image_ref,
            image_auth=(image_auth_user, image_auth_password),
            core_host=core_host,
            core_auth=(
                core_user, core_token) if core_user and core_token else None,
        )

        manifest_manager = ManifestManager()
        if entry := manifest_manager.query("dependencies", package_name):
            if entry.get("installer") != "image":
                raise Exception(
                    f"Package {package_name} already exists with different installer."
                    "\nPossible solutions:"
                    "\n\t- Remove package from manifest and try again. Use `malevich remove` command"  # noqa: E501
                    "\n\t- Use `malevich restore` command to restore package with its installer"  # noqa: E501
                )

            manifest_manager.put(
                "dependencies",
                package_name,
                value=manifest_entry.model_dump(),
            )
            if progress:
                progress.update(
                    task_,
                    description="\n[green]✔[/green] Package "
                    f"[blue]({package_name})[/blue] "
                    "[yellow]updated[/yellow] successfully",
                    completed=1,
                )
        else:
            if progress:
                progress.update(
                    task_,
                    description="\n[green]✔[/green] Package "
                    f"[blue]({package_name})[/blue] "
                    "installed successfully",
                    completed=1,
                )
            manifest_manager.put(
                "dependencies",
                value={f"{package_name}": manifest_entry.model_dump()},
                append=True,
            )

        manifest_manager.save()
        return True
    except Exception as err:
        if progress:
            progress.update(
                task_,
                description="\n[red]✘[/red] Failed with "
                "[yellow]image[/yellow] installer "
            )
        raise err

def underscored(entry: str):
    return re.sub(r"[\W\s]", "_", entry)

def _install_from_space(
    package_name: str,
    reverse_id: str,
    branch: Optional[str] = None,
    version: Optional[str] = None,
    progress: Progress = None,
) -> None:
    installer = SpaceInstaller()
    flow_installer = FlowInstaller()
    if progress:
        task_ = progress.add_task(
            f"Attempting to install [b blue]{package_name}[/b blue] with"
            " [i yellow]space[/i yellow] installer",
            total=1,
        )

    try:
        manifest_entry = installer.install(
            package_name=package_name,
            reverse_id=reverse_id,
            branch=branch,
            version=version,
        )
        manifest_manager = ManifestManager()

        if isinstance(manifest_entry, LoadedComponentSchema):
                flow_branch_version, active_branch, active_versions = Space.fetch(
                    reverse_id=reverse_id,
                )
                integrations = []
                for branch, versions in flow_branch_version.items():
                    for version, uid in versions.items():
                        task = Space(
                            reverse_id=reverse_id,
                            uid=uid,
                            policy='fetch',
                        )

                        injectables = task.get_injectables()

                        mappings = {}
                        for col in injectables:
                            mappings[underscored(col.alias)] = col.alias

                        integrations.append(
                            Integration(
                                mapping=mappings,
                                version=version,
                                deployment=uid,
                                injectables=injectables,
                                branch=branch
                            )
                        )

                flow_installer.install(reverse_id=reverse_id, integrations=integrations)
                manifest_manager.put('flows', reverse_id, value=integrations)
                if progress:
                    progress.update(
                        task_,
                        description="\n[green]✔[/green] Flow "
                        f"[blue]({reverse_id})[/blue] "
                        "synced successfully",
                        completed=1,
                    )

        else:
            if entry := manifest_manager.query("dependencies", package_name):
                if entry.get("installer") != "space":
                    raise Exception(
                        f"Package {package_name} already exists "
                        "with different installer."
                        "\nPossible solutions:"
                        "\n\t- Remove package from manifest and try again. Use `malevich remove` command"  # noqa: E501
                        "\n\t- Use `malevich restore` command to restore package with its installer"  # noqa: E501
                    )

                manifest_manager.put(
                    "dependencies",
                    package_name,
                    value=manifest_entry.model_dump(),
                )
                if progress:
                    progress.update(
                        task_,
                        description="\n[green]✔[/green] Package "
                        f"[blue]({package_name})[/blue] "
                        "[yellow]updated[/yellow] successfully",
                        completed=1,
                    )
            else:
                if progress:
                    progress.update(
                        task_,
                        description="\n[green]✔[/green] Package "
                        f"[blue]({package_name})[/blue] "
                        "installed successfully",
                        completed=1,
                    )
                manifest_manager.put(
                    "dependencies",
                    value={f"{package_name}": manifest_entry.model_dump()},
                    append=True,
                )

        manifest_manager.save()
        return True
    except Exception as err:
        if progress:
            progress.update(
                task_,
                description="\n[red]✘[/red] Failed with "
                "[yellow]space[/yellow] installer "
            )
        raise err

def _install_from_local(
    package_name: str,
    import_path: str,
    progress: Progress = None,
) -> None:
    installer = LocalInstaller()
    if progress:
        task_ = progress.add_task(
            f"Attempting to install [b blue]{package_name}[/b blue] with"
            " [i yellow]local[/i yellow] installer",
            total=1,
        )

    try:
        manifest_entry = installer.install(
            package_name=package_name,
            app_path=Path(import_path)
        )
        manifest_manager = ManifestManager()
        manifest_manager.put(
            "dependencies",
            package_name,
            value=manifest_entry.model_dump(),
        )
        manifest_manager.save()
        if progress:
            progress.update(
                task_,
                description="\n[green]✔[/green] Package "
                f"[blue]({package_name})[/blue] "
                "installed successfully",
                completed=1,
            )
        return True
    except Exception as err:
        if progress:
            progress.update(
                task_,
                description="\n[red]✘[/red] Failed with "
                "[yellow]local[/yellow] installer "
            )
        raise err

@use.command("image", help=help['image --help'])
def install_from_image(
    package_name: Annotated[
        str,
        typer.Argument(help="Package name")
    ],
    image_ref: Annotated[
        Optional[str],
        typer.Argument(
            help="Image reference in format <registry>/<image>:<tag>"
        )
    ] = None,
    image_auth_user: Annotated[
        str,
        typer.Argument(
            help="Image registry auth user"
        )
    ] = "",
    image_auth_password: Annotated[
        str,
        typer.Argument(
            help="Image registry auth password"
        )
    ] = "",
    core_host: Annotated[
        str,
        typer.Option(
            help="Malevich Core hostname"
        )
    ] = DEFAULT_CORE_HOST,
    core_user: Annotated[
        Optional[str],
        typer.Option(
            help="Malevich Core user"
        )
    ] = None,
    core_token: Annotated[
        Optional[str],
        typer.Option(
            help="Malevich Core token"
        )
    ] = None,
) -> None:
    try:
        if image_ref is None:
            return _install_from_image(
                package_name=package_name,
                image_ref=IMAGE_BASE.format(app=package_name),
                image_auth_user=image_auth_user,
                image_auth_password=image_auth_password,
                core_host=core_host or DEFAULT_CORE_HOST,
                core_user=core_user,
                core_token=core_token,
                progress=Progress(
                    SpinnerColumn(), TextColumn("{task.description}")),
            )
        else:
            return _install_from_image(
                package_name=package_name,
                image_ref=image_ref,
                image_auth_user=image_auth_user,
                image_auth_password=image_auth_password,
                core_host=core_host or DEFAULT_CORE_HOST,
                core_user=core_user,
                core_token=core_token,
                progress=Progress(
                    SpinnerColumn(), TextColumn("{task.description}")),
            )
    except Exception as err:
        rich.print("\n\n[red]Installation failled[/red]")
        rich.print(err)
        raise err
        exit(-1)


@use.command("space", help=help['space --help'])
def install_from_space(
    package_name: Annotated[
        str,
        typer.Argument(help="Package name")
    ],
    reverse_id: Annotated[
        str,
        typer.Argument(help="Reverse id of the component")
    ],
    deployment_id: str=typer.Option(
        None,
        '--deployment-id',
        '-d',
        show_default=False,
        help='Flow Deployment ID. If not set, will get active version flow.'
    ),
    branch: str=typer.Option(
        None,
        '--branch',
        '-b',
        show_default=False,
        help="Flow branch. If not specified, will take the active one."
    ),
    attach_any: bool=typer.Option(
        False,
        '--attach_any',
        '-a',
        show_default=False,
        help="Attach to any flow deployment"
    ),
    version: Annotated[
        Optional[str],
        typer.Argument(help="Version of the component")
    ] = None,
) -> None:
    try:
        return _install_from_space(
            package_name=package_name,
            reverse_id=reverse_id,
            branch=branch,
            version=version,
            attach_any=attach_any,
            deployment_id=deployment_id,
            progress=Progress(
                SpinnerColumn(), TextColumn("{task.description}")),
        )
    except Exception as err:
        rich.print("\n\n[red]Installation failled[/red]")
        rich.print(err)
        exit(-1)


@use.command("local", help=help['local --help'])
def install_from_local(
    package_name: Annotated[
        str,
        typer.Argument(help="Package name")
    ],
    path: Annotated[
        str,
        typer.Argument(help="Path to the package")
    ],
) -> None:
    try:
        return _install_from_local(
            package_name=package_name,
            import_path=path,
            progress=Progress(
                SpinnerColumn(), TextColumn("{task.description}")),
        )
    except Exception as err:
        raise err
        rich.print("\n\n[red]Installation failled[/red]")
        rich.print(err)
        exit(-1)




import hashlib
from typing import Any, Optional

from pydantic import BaseModel

from ...constants import DEFAULT_CORE_HOST
from ..dependency import Dependency
from .compat import CompatabilityStrategy, compare_images
from .space import SpaceDependency


class ImageOptions(BaseModel):
    image_ref: Optional[str] = None
    image_auth_user: Optional[str] = None
    image_auth_pass: Optional[str] = None
    core_host: Optional[str] = DEFAULT_CORE_HOST
    core_auth_user: Optional[str] = None
    core_auth_token: Optional[str] = None


class ImageDependency(Dependency):
    installer: str = 'image'
    options: Optional[ImageOptions] = ImageOptions()

    def simple(self) -> dict[str, Any]:
        simpler = {**self.options.model_dump()}
        return {k: v for k, v in simpler.items() if v is not None}

    def checksum(self) -> str:
        return hashlib.sha256(self.model_dump_json().encode()).hexdigest()

    def compatible_with(
        self,
        other: 'Dependency',
        compatability_strategy: CompatabilityStrategy = CompatabilityStrategy()
    ) -> bool:
        if isinstance(other, SpaceDependency) or isinstance(other, ImageDependency):
            return compare_images(
                self.options.image_ref,
                other.options.image_ref,
                compatability_strategy
            )
        else:
            return False



import sys
from functools import cache
from pathlib import Path


@cache
def get_malevich_version():
    module_path = Path(sys.modules['malevich'].__file__)
    malevich_version = open(module_path.parent.parent / 'VERSION').read().strip()
    return malevich_version


import json
import re
from pathlib import Path

from malevich_coretools import Condition, JsonImage, Processor, Result

from malevich._autoflow.tracer import traced
from malevich._utility import unique
from malevich._utility.logging import LogLevel, cout
from malevich._utility.registry import Registry
from malevich.interpreter.abstract import Interpreter
from malevich.interpreter.core import CoreInterpreter
from malevich.models import BaseNode
from malevich.models.actions import Action
from malevich.models.nodes.asset import AssetNode
from malevich.models.nodes.collection import CollectionNode
from malevich.models.nodes.document import DocumentNode
from malevich.models.nodes.operation import OperationNode
from malevich.models.preferences import VerbosityLevel
from malevich.models.state.local import LocalInterpreterState
from malevich.models.task.interpreted.local import LocalTask

_levels = [
    LogLevel.Info,
    LogLevel.Warning,
    LogLevel.Error,
    LogLevel.Debug
]

_actions = [
    Action.Interpretation,
    Action.Preparation,
    Action.Run,
    Action.Results
]

def _log(
    message: str,
    level: int = 0,
    action: int = 0,
    step: bool = False,
    *args,
) -> None:
    cout(
        _actions[action],
        message,
        verbosity=VerbosityLevel.AllSteps if step else VerbosityLevel.OnlyStatus,
        level=_levels[level],
        *args,
    )

registry = Registry()

class LocalInterpreter(CoreInterpreter):
    def __init__(self) -> None:
        super(CoreInterpreter, self).__init__(LocalInterpreterState())
        self._state = LocalInterpreterState()
        self.update_state()

    def before_interpret(self, state: LocalInterpreterState) -> LocalInterpreterState:
        _log("Interpretation in local mode", 0, 0, True)
        return state

    def create_node(
        self,
        state: LocalInterpreterState,
        tracer: traced[BaseNode]
    ) -> LocalInterpreterState:
        match tracer.owner:
            case OperationNode():
                if tracer.owner.alias is None:
                    tracer.owner.alias = unique(tracer.owner.processor_id)
                
                state.operation_nodes[tracer.owner.alias] = tracer.owner

                extra = registry.get(tracer.owner.operation_id)
                if extra is not None:
                    if 'package_path' not in extra:
                        raise ValueError(
                            f'Processor {tracer.owner.processor_id} has '
                            'installed with installer different from "local". '
                            'Using local mode is only available for local '
                            'installations.'
                        )

                    package_path = Path(extra['package_path'])

                    if not package_path.exists():
                        raise FileNotFoundError(
                            f'Processor {tracer.owner.processor_id} has '
                            f'been installed from path {package_path}, which '
                            'does not exist.'
                        )

                else:
                    raise ValueError(
                        f'Failed to find processor {tracer.owner.processor_id} '
                        'in the registry.'
                    )

                state.import_paths.add(str(package_path.absolute()))

                if tracer.owner.is_condition:
                    state.conditions[tracer.owner.alias] = Condition(
                        arguments={},
                        cfg=json.dumps(tracer.owner.config),
                        image=JsonImage(ref='', user='', token=''),
                        conditionId=tracer.owner.processor_id
                    )
                else:
                    state.processors[tracer.owner.alias] = Processor(
                        arguments={},
                        cfg=json.dumps(tracer.owner.config),
                        image=JsonImage(ref=''),
                        processorId=tracer.owner.processor_id
                    )

                state.results[tracer.owner.alias] = [Result(name=tracer.owner.alias)]

            case CollectionNode():
                if tracer.owner.alias is None:
                    tracer.owner.alias = unique(tracer.owner.collection.collection_id)
                state.collection_nodes[tracer.owner.alias] = tracer.owner
            case AssetNode():
                tracer.owner.alias = unique(tracer.owner.name)
                state.asset_nodes[tracer.owner.alias] = tracer.owner
            case DocumentNode():
                if tracer.owner.alias is None:
                    tracer.owner.alias = unique(tracer.owner.reverse_id)
                state.document_nodes[tracer.owner.alias] = tracer.owner
            case _:
                pass

        _log(f"Node: {tracer.owner.uuid}, {tracer.owner.short_info()}", -1, 0, True)
        return state

    def get_task(self, state: LocalInterpreterState) -> LocalTask:
        return LocalTask(state)

    def attach(
        self,
        unique_task_hash: str,
        only_fetch: bool = False,
    ) -> LocalTask:
        raise NotImplementedError("Local interpreter does not support attach method")



"""
Provides functionality to install packages from plain Docker images
using Malevich Core capabilities
"""

import hashlib
import re
from typing import Optional

from malevich._utility import Stub
from malevich.constants import DEFAULT_CORE_HOST, IMAGE_BASE
from malevich.install import Installer
from malevich.manifest import ManifestManager
from malevich.models import ImageDependency, ImageOptions
from malevich.path import Paths

from .._core.scan import scan_core

_pydantic_types = {
    "string": str.__name__,
    "integer": int.__name__,
    "number": float.__name__,
    "boolean": bool.__name__,
}


class Templates:
    """Templates for creating a package from a Docker image"""

    disclaimer = """\"\"\"
THIS FILE IS AUTOGENERATED BY malevich PACKAGE INSTALLER.

THIS FILE CONTAINS VITAL INFORMATION ABOUT THE PACKAGE AND ITS CONTENTS.
DO NOT MODIFY THIS FILE MANUALLY.
\"\"\"
"""

    imports = """
from malevich._autoflow.function import autotrace, sinktrace
from malevich._utility import Registry
from malevich.models.nodes import OperationNode

from pydantic import BaseModel
"""
    registry = """
Registry().register("{operation_id}", {{
    "image_ref": {image_ref},
    "image_auth_user": {image_auth_user},
    "image_auth_pass": {image_auth_pass},
    "processor_id": "{processor_id}",
}})
"""

    schema = """
class {schema_name}(BaseModel):
{schema_def}
"""

    processor = """
@{decor}
def {name}({args}config: dict = {{}}):
    \"\"\"{docs}\"\"\"
    return OperationNode(
        operation_id="{operation_id}",
        config=config,
        processor_id="{name}",
        package_id="{package_id}",
        alias=alias,
    )
"""
    init = """
# {init_name} = ...

"""

    input = """
# {input_name} = ...

"""

    output = """
# {output_name} = ...

"""


class ImageInstaller(Installer):
    """Installs package using Docker image"""
    name = "image"


    def install(
        self,
        package_name: str,
        image_ref: str,
        image_auth: tuple[str, str],
        core_host: str = DEFAULT_CORE_HOST,
        core_auth: Optional[tuple[str, str]] = None,
    ) -> ImageDependency:
        package_name = re.sub(r'[\W\s]+', '_', package_name)
        app_info = scan_core(
            core_auth=core_auth,
            core_host=core_host,
            image_ref=image_ref,
            image_auth=image_auth,
        )
        checksum = hashlib.sha256(
            app_info.model_dump_json().encode()).hexdigest()
        m = ManifestManager()
        iauth_user, iauth_pass, cauth_user, cauth_token = m.put_secrets(
            image_auth_user=image_auth[0],
            image_auth_password=image_auth[1],
            core_auth_user=core_auth[0] if core_auth else None,
            core_auth_token=core_auth[1] if core_auth else None,
            salt=checksum,
        )
        dependency = ImageDependency(
            package_id=package_name,
            version="",
            installer="image",
            options=ImageOptions(
                core_host=core_host,
                core_auth_token=cauth_token,
                core_auth_user=cauth_user,
                image_auth_user=iauth_user,
                image_auth_pass=iauth_pass,
                image_ref=image_ref,
            ),
        )

        operation_ids = {
            str(op.id): hashlib.sha256(op.model_dump_json().encode()).hexdigest()
            for op in [*app_info.processors.values(), *app_info.conditions.values()]
        }

        operation_names = {
            str(op.id): op.name
            for op in [*app_info.processors.values(), *app_info.conditions.values()]
        }

        Stub.from_app_info(
            app_info=app_info,
            path=Paths.module(package_name),
            package_name=package_name,
            dependency=dependency,
            operation_ids=operation_ids,
            registry_records={
                processor_id: {
                    "operation_id": operation_id,
                    "image_ref":  ('dependencies', package_name, 'options', 'image_ref'),  # noqa: E501
                    "image_auth_user": ('dependencies', package_name, 'options', 'image_auth_user'),  # noqa: E501
                    "image_auth_pass": ('dependencies', package_name, 'options', 'image_auth_pass'),  # noqa: E501
                    "processor_id": operation_names[processor_id],
                }
                for processor_id, operation_id in operation_ids.items()
            },
        )

        return dependency

    def restore(self, dependency: ImageDependency) -> ImageDependency:
        core_user = dependency.options.core_auth_user
        core_token = dependency.options.core_auth_token
        image_ref = dependency.options.image_ref
        if not image_ref:
            image_ref = IMAGE_BASE.format(app=dependency.package_id)

        return self.install(
            package_name=dependency.package_id,
            image_ref=image_ref,
            image_auth=(
                dependency.options.image_auth_user or "",
                dependency.options.image_auth_pass or ""
            ),
            core_host=dependency.options.core_host,
            core_auth=(
                core_user,
                core_token,
            ) if core_user and core_token else None,
        )

    def construct_dependency(self, object: dict) -> ImageDependency:
        manf = ManifestManager()
        parsed = ImageDependency(**object)
        parsed.options.image_auth_pass = manf.query_secret(
            parsed.options.image_auth_pass,
            only_value=True,
        )
        parsed.options.image_auth_user = manf.query_secret(
            parsed.options.image_auth_user,
            only_value=True,
        )
        parsed.options.core_auth_user = manf.query_secret(
            parsed.options.core_auth_user,
            only_value=True,
        )
        parsed.options.core_auth_token = manf.query_secret(
            parsed.options.core_auth_token,
            only_value=True,
        )
        return parsed



import typing as _typing

from .base import BaseResult
from .core import *
from .space import *

# Possible result types
Result = _typing.Union[
    BaseResult,
    CoreLocalDFResult,
    CoreResult,
    SpaceCollectionResult
]


__all__ = [
    "BaseResult",
    "CoreLocalDFResult",
    "CoreResult",
    "CoreResultPayload",
    "SpaceCollectionResult",
    "Result",
]




import json
import logging
import pickle
from typing import Any, Dict, Generic, List, Optional, Tuple, TypeVar, Union

import boto3
import jsonpickle
import numpy as np
import pandas as pd
from botocore.response import StreamingBody

from .df import OBJ

WORKDIR = "/malevich"
"""
Working directory from which the app is run.
Equivalent to :code:`os.getcwd()` from within the app.
"""

APP_DIR = f"{WORKDIR}/apps"
"""
Directory into which the user code is copied during app construction.
"""

MinimalCfg = TypeVar('MinimalCfg')

class Context(Generic[MinimalCfg]):
    """
    Context contains all the necessary information about the run
    and the environment in which it is executed. Also, context provides
    a number of auxiliary functions  for interacting with the environment, such as
    working with shared storage (:meth:`share`, :meth:`get_share_path`, :meth:`delete_share`),
    dealing with common objects (:attr:`common`),
    access to the key-value storage (:attr:`dag_key_value`),
    and object storage (:attr:`object_storage`).
    """

    class _DagKeyValue:
        """
        Simple key-value storage, shared for all apps of one run.
        Values must be bytes, string, int or float; dictionary order is not guaranteed
        """

        def __init__(self, run_id: Optional[str] = None) -> None:
            pass

        def get_bytes(self, key: str) -> bytes:
            """Gets a binary value by key more optimally.

            Consider using this to retrieve binary data in
            more efficient way, but keep in mind that :meth:`get`
            can also be used

            Args:
                key (str): key in storage

            Returns:
                bytes: Value stored by key
            """
            pass

        async def async_get_bytes(self, key: str) -> bytes:
            """Gets a binary value by key more optimally.

            Consider using this to retrieve binary data in
            more efficient way, but keep in mind that :meth:`get`
            can also be used

            Args:
                key (str): key in storage

            Returns:
                bytes: Value stored by key
            """
            pass

        def get(self, keys: List[str]) -> Dict[str, Any]:
            """Gets values by keys

            Retrieves a slice of storage by keys. If a key is not found,
            the value will be set to None

            Args:
                keys (list[str]): list of keys

            Returns:
                Dict[str, Any]: A dictionary of key-value pairs.
            """
            pass

        async def async_get(self, keys: List[str]) -> Dict[str, Any]:
            """Gets values by keys

            Retrieves a slice of storage by keys. If a key is not found,
            the value will be set to None

            Args:
                keys (list[str]): list of keys

            Returns:
                Dict[str, Any]: A dictionary of key-value pairs.
            """
            pass

        def get_all(self) -> Dict[str, Any]:
            """Gets all values

            Retrieves the whole storage as a dictionary.

            Returns:
                Dict[str, Any]: A dictionary of key-value pairs.
            """
            pass

        async def async_get_all(self) -> Dict[str, Any]:
            """Gets all values

            Retrieves the whole storage as a dictionary.

            Returns:
                Dict[str, Any]: A dictionary of key-value pairs.
            """
            pass

        def update(self, keys_values: Dict[str, Any]) -> None:
            """Sets values by keys

            Accepts a dictionary of key-value pairs and sets them in storage.
            If a key already exists, it will be overwritten.

            Args:
                keys_values (dict): A dictionary of key-value pairs.
            """
            pass

        async def async_update(self, keys_values: Dict[str, Any]) -> None:
            """Sets values by keys

            Accepts a dictionary of key-value pairs and sets them in storage.
            If a key already exists, it will be overwritten.

            Args:
                keys_values (dict): A dictionary of key-value pairs.
            """
            pass

        def clear(self) -> None:
            """Purges the storage"""
            pass

        async def async_clear(self) -> None:
            """Purges the storage"""
            pass

    class _ObjectStorage:
        """
        A storage for binary objects common to the user.

        Works with cloud S3 storage and app shared file system. Provides
        an access to the objects shared with :meth:`Context.share` method.
        and accessible with :meth:`Context.get_share_path` method.

        All methods of the object storage have :code:`local` parameter. If it is
        set to :code:`True`, the operation will be performed on the local file system.
        Otherwise, the operation will be performed on the remote object storage.

        The information might be unsynchronized between apps. To ensure that the
        information is synchronized, use :code:`all_apps` parameter. If it is set to
        :code:`True`, the operation will be performed on all apps. Otherwise, the
        operation will be performed only on apps the method is called from.

        """

        def get_keys(self, local: bool = False, all_apps: bool = False) -> List[str]:
            """
            Get keys from local mount or remote object storage.

            Args:
                local (bool, optional): whether to use local mount or remote object storage.
                    If set to :code:`True`, the operation will
                    be performed on the local file system. Otherwise,
                    the operation will be performed on the remote object storage. Defaults to False.
                all_apps (bool, optional): whether to synchronize the operation between all apps.
                    If set to :code:`True`, the operation will be performed on all apps.
                    Otherwise, the operation will be performed only on apps the method is called from.
                    Defaults to False.

            Returns:
                List[str]: Keys from local mount or remote object storage.
            """
            pass

        async def async_get_keys(self, local: bool = False, all_apps: bool = False) -> List[str]:  # noqa: E501
            """
            Get keys from local mount or remote object storage.

            Args:
                local (bool, optional): whether to use local mount or remote object storage.
                    If set to :code:`True`, the operation will
                    be performed on the local file system. Otherwise,
                    the operation will be performed on the remote object storage. Defaults to False.
                all_apps (bool, optional): whether to synchronize the operation between all apps.
                    If set to :code:`True`, the operation will be performed on all apps.
                    Otherwise, the operation will be performed only on apps the method is called from.
                    Defaults to False.

            Returns:
                List[str]: Keys from local mount or remote object storage.
            """
            pass

        def get(self, keys: List[str], force: bool = False, all_apps: bool = True) -> List[str]:    # noqa: E501
            """
            Updates mount for this app (or all apps), return keys for which it was successful

            Args:
                keys (List[str]):
                    Keys by which values are obtained.
                    If this is not possible, this key will not be returned in result)
                force (bool, optional):
                    If set, it will ignore what is locally and
                    load data from the remote object storage.
                    Otherwise it will only take what does not exist.
                    Defaults to False.
                all_apps (bool, optional):
                    If set to true, the operation will be performed in all apps.
                    Otherwise only for apps with associated mount.
                    Defaults to True.

            Returns:
                List[str]: Keys by which it was possible
                to obtain the value and load it into the mount
            """
            pass

        async def async_get(self, keys: List[str], force: bool = False, all_apps: bool = True) -> List[str]:    # noqa: E501
            """
            Updates mount for this app (or all apps), return keys for which it was successful

            Args:
                keys (List[str]):
                    Keys by which values are obtained.
                    If this is not possible, this key will not be returned in result)
                force (bool, optional):
                    If set, it will ignore what is locally and
                    load data from the remote object storage.
                    Otherwise it will only take what does not exist.
                    Defaults to False.
                all_apps (bool, optional):
                    If set to true, the operation will be performed in all apps.
                    Otherwise only for apps with associated mount.
                    Defaults to True.

            Returns:
                List[str]: Keys by which it was possible
                to obtain the value and load it into the mount
            """
            pass

        def get_all(
            self,
            local: bool = False,
            force: bool = False,
            all_apps: bool = True
        ) -> List[str]:
            """
            Updates mount and return all keys in it,
            if `local` - return result only for mount (or all apps mounts if `all_apps`),
            otherwise - load all by remote object storage

            Args:
                local (bool, optional): whether to use local mount or remote object storage.
                    If set to :code:`True`, the operation will
                    be performed on the local file system. Otherwise,
                    the operation will be performed on the remote object storage. Defaults to False.
                force (bool, optional):
                    If set, it will ignore what is locally and
                    load data from the remote object storage.
                    Otherwise it will only take what does not exist.
                    Defaults to False.
                all_apps (bool, optional):
                    If set to true, the operation will be performed in all apps.
                    Otherwise only for apps with associated mount.
                    Defaults to True.

            Returns:
                List[str]: All keys in the mount or all apps mounts if `all_apps` is True,
                otherwise load all keys from remote object storage.
            """
            pass

        async def async_get_all(
            self,
            local: bool = False,
            force: bool = False,
            all_apps: bool = True
        ) -> List[str]:
            """
            Updates mount and return all keys in it,
            if `local` - return result only for mount (or all apps mounts if `all_apps`),
            otherwise - load all by remote object storage

            Args:
                local (bool, optional): whether to use local mount or remote object storage.
                    If set to :code:`True`, the operation will
                    be performed on the local file system. Otherwise,
                    the operation will be performed on the remote object storage. Defaults to False.
                force (bool, optional):
                    If set, it will ignore what is locally and
                    load data from the remote object storage.
                    Otherwise it will only take what does not exist.
                    Defaults to False.
                all_apps (bool, optional):
                    If set to true, the operation will be performed in all apps.
                    Otherwise only for apps with associated mount.
                    Defaults to True.

            Returns:
                List[str]: All keys in the mount or all apps mounts if `all_apps` is True,
                otherwise load all keys from remote object storage.
            """
            pass

        def update(
            self,
            keys: List[str],
            presigned_expire: Optional[int] = -1
        ) -> Dict[str, str]:
            """Updates objects in remote storage

            Retrieves objects from local mount and updates remote object storage.
            If :code:`presigned_expire` is set to a positive value, creates
            and returns presigned urls for the objects.

            If :code:`presigned_expire` is None, set to default timeout.

            If :code:`presigned_expire` is negative, returns an empty dictionary.

            Args:
                keys (List[str]): Keys to update
                presigned_expire (int, optional):
                    If positve, life span of presigned urls in seconds.
                    If None, set to default timeout.
                    Defaults to -1.

            Returns:
                Dict[str, str]: Mapping of keys to presigned urls
            """
            pass

        async def async_update(
            self,
            keys: List[str],
            presigned_expire: Optional[int] = -1
        ) -> Dict[str, str]:
            """Updates objects in remote storage

            Retrieves objects from local mount and updates remote object storage.
            If :code:`presigned_expire` is set to a positive value, creates
            and returns presigned urls for the objects.

            If :code:`presigned_expire` is None, set to default timeout.

            If :code:`presigned_expire` is negative, returns an empty dictionary.

            Args:
                keys (List[str]): Keys to update
                presigned_expire (int, optional):
                    If positve, life span of presigned urls in seconds.
                    If None, set to default timeout.
                    Defaults to -1.

            Returns:
                Dict[str, str]: key to presigned url
            """
            pass

        def presigned(
            self,
            keys: List[str],
            expire: Optional[int] = None
        ) -> Dict[str, str]:
            """Creates presigned urls for specified keys

            Args:
                keys (List[str]): Keys to create presigned urls for
                expire (int, optional):
                    Life span of presigned urls in seconds (must be positive).
                    If None, set to default timeout.
                    Defaults to None.
            Returns:
                Dict[str, str]: Mapping of keys to presigned urls
            """
            pass

        async def async_presigned(
            self,
            keys: List[str],
            expire: Optional[int] = None
        ) -> Dict[str, str]:
            """Creates presigned urls for specified keys

            Args:
                keys (List[str]): Keys to create presigned urls for
                expire (int, optional):
                    Life span of presigned urls in seconds (must be positive).
                    If None, set to default timeout.
                    Defaults to None.
            Returns:
                Dict[str, str]: Mapping of keys to presigned urls
            """
            pass

        def delete(self, keys: List[str]) -> None:
            """Deletes values in mount and remote storage

            Args:
                keys (List[str]): Keys to delete
            """
            pass

        async def async_delete(self, keys: List[str]) -> None:
            """Deletes values in mount and remote storage

            Args:
                keys (List[str]): Keys to delete
            """
            pass

    app_id: str
    """
    App ID (unique for each app).
    This ID given to an app at startup by interpreters.
    """

    run_id: str
    """
    Run ID (unique for each run).
    """

    app_cfg: Dict[str, Any]
    """
    App configuration. Given to the app at startup
    and contains arbitrary configuration data. The configuration
    is preserved between runs. This field is used to
    dictate the behaviour of the app.
    """

    dag_key_value: _DagKeyValue
    """
    Key-value storage. Shared between all apps of one run.
    Values must be bytes, string, int or float; dictionary order is not guaranteed
    """

    object_storage: _ObjectStorage
    """
    Object storage, shared between all apps of one run.
    """

    common: Any
    """
    An object shared between all runs of the app.
    It is not being explicitly serialized, so may be any object.
    In comparison, objects put into :attr:`app_cfg` can be used
    in the same way, but they are serialized and deserialized limiting
    the types that can be used. That is not the case for :attr:`common`.
    """

    def __init__(self) -> None:
        self.app_id: str = ""                                       # app id at startup
        self.run_id: str = ""                                       # run id at startup
        self.app_cfg: Union[MinimalCfg, Dict[str, Any]] = {}        # configuration given to the app at startup  # noqa: E501
        self.msg_url: str = ""                                      # default url for msg operation              # noqa: E501
        self.email: Optional[str] = None                            # email for email_send operation             # noqa: E501
        self.dag_key_value = Context._DagKeyValue(
            self.run_id)      # key-value storage
        self.object_storage = Context._ObjectStorage()              # object storage
        self.common = None                                          # arbitrary common variable between app runs # noqa: E501
        self.logger = logging.getLogger(f"{self.operation_id}${self.run_id}")

    def share(
        self,
        path: str,
        all_runs: bool = False,
        path_prefix: str = APP_DIR,
        force: bool = False,
        synchronize: bool = True
    ) -> None:
        """Shares a file or a directory between all apps for the current deployment.

        The file or directory is copied to a shared directory, which is
        accessible by all apps. The file should be accessible by the
        following path:

            :code:`{path_prefix}/{path}`

        where :code:`path_prefix` and :code:`path` are the corresponding arguments.

        -----------------------------------------------------------------------------

        Note, that:

        *   The :code:`path_prefix` is not included into shared key.
        *   If :code:`all_runs` is set to :code:`True`,
            the file or directory will be shared between all runs of the app.
            Otherwise, it will be shared only between the apps of the current run.
        *   If :code:`synchronize` is set to :code:`True`,
            the file or directory will be copied to all apps.
            Otherwise, it will be copied only to apps with a common mount.
        *   If the file or directory already exists in the shared directory,
            it will not be copied unless :code:`force` is set to :code:`True`.

        -----------------------------------------------------------------------------

        .. note::

            See `Sharing files <#sharing-files>`_ for more details.

        Args:
            path (str):
                Key of shared file or directory
            all_runs (bool, optional):
                Whether to share the file or directory between all runs.
                Defaults to False.
            path_prefix (str, optional):
                Path prefix.
                Defaults to `APP_DIR`.
            force (bool, optional):
                Whether to overwrite the file or directory if it already exists.
                Defaults to False.
            synchronize (bool, optional):
                Whether to synchronize the file or directory between all apps.
                Defaults to True.
        """
        pass

    async def async_share(
        self,
        path: str,
        all_runs: bool = False,
        path_prefix: str = APP_DIR,
        force: bool = False,
        synchronize: bool = True
    ) -> None:
        """Shares a file or a directory between all apps for the current deployment.

        It is an asynchronous version of :meth:`share`.

        Args:
            path (str):
                Key of shared file or directory
            all_runs (bool, optional):
                Whether to share the file or directory between all runs.
                Defaults to False.
            path_prefix (str, optional):
                Path prefix.
                Defaults to `APP_DIR`.
            force (bool, optional):
                Whether to overwrite the file or directory if it already exists.
                Defaults to False.
            synchronize (bool, optional):
                Whether to synchronize the file or directory between all apps.
                Defaults to True.
        """
        pass

    def share_many(
        self,
        paths: List[str],
        all_runs: bool = False,
        path_prefix: str = APP_DIR,
        force: bool = False,
        synchronize: bool = True
    ) -> None:
        """Shares multiple files or directories.

        The same as :meth:`share`, but for multiple files or directories.
        Ignores paths that do not exist.
        """

    async def async_share_many(
        self,
        paths: List[str],
        all_runs: bool = False,
        path_prefix: str = APP_DIR,
        force: bool = False,
        synchronize: bool = True
    ) -> None:
        """Shares multiple files or directories.
        The same as :meth:`async_share`, but for multiple files or directories.
        Ignores paths that do not exist.
        """

    def get_share_path(
        self,
        path: str,
        all_runs: bool = False,
        not_exist_ok: bool = False
    ) -> str | None:
        """Retrieves a real file path by shared key.

        Once file is shared, it is copied to a shared directory, which is
        accessible by all apps. However, to actually access the file, you
        have to retrieve its real path using in the mounted file system.

        .. note::

            See `Accessing shared files <#accessing-shared-files>`_ for more details.


        Args:
            path (str): Shared file key
            all_runs (bool, optional):
                To access files shared between all runs, set to :code:`True`.
                Defaults to False.
            not_exist_ok (bool, optional):
                Whether to raise an exception if the file does not exist.

        Returns:
            str: Readable file path or None (if :code:`not_exist_ok` is set to :code:`True`)
        """
        pass

    def delete_share(
        self,
        path: str,
        all_runs: bool = False,
        synchronize: bool = True
    ) -> None:
        """Deletes previously shared file or directory.

        Args:
            path (str): Shared file key
            all_runs (bool, optional):
                Whether to delete files shared between all runs.
                Defaults to False.
            synchronize (bool, optional):
                Whether to synchronize the deletion between all apps.
                Defaults to True.
        """
        pass

    async def async_delete_share(
        self, path: str, all_runs: bool = False, synchronize: bool = True
    ) -> None:
        """Deletes previously shared file or directory.

        Args:
            path (str): Shared file key
            all_runs (bool, optional):
                Whether to delete files shared between all runs.
                Defaults to False.
            synchronize (bool, optional):
                Whether to synchronize the deletion between all apps.
                Defaults to True.
        """
        pass

    def synchronize(
        self,
        paths: Optional[List[str]] = None,
        all_runs: bool = False
    ) -> None:  # TODO synchronize removing
        """Forcefully synchronizes paths accross all apps.

        Setting :code:`paths` to :code:`None` or an empty list will synchronize
        all shared files and directories.

        Args:
            paths (Optional[List[str]], optional): Paths to synchronize.
            all_runs (bool, optional):
                Whether to synchronize files shared between all runs.
                Should be set to the same value as :meth:`share` method was called with.
                Defaults to False.
        """
        pass

    async def async_synchronize(
        self, paths: Optional[List[str]] = None, all_runs: bool = False
    ) -> None:
        """Forcefully synchronizes paths accross all apps.

        Setting :code:`paths` to :code:`None` or an empty list will synchronize
        all shared files and directories.

        Args:
            paths (Optional[List[str]], optional): Paths to synchronize.
            all_runs (bool, optional):
                Whether to synchronize files shared between all runs.
                Should be set to the same value as :meth:`share` method was called with.
                Defaults to False.
        """
        pass

    def msg(
        self,
        data: Union[str, Dict],
        url: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
        wait: bool = False,
        wrap: bool = True,
        with_result: bool = False
    ) -> None:
        """Sends http msg to system or any url

        Args:
            data (Union[str, Dict]): msg data
            url (Optional[str]): url to which the request is sent. If not specified, the default url is used (system url, but it can be overridden in the startup cfg, the parameter is msg_url). system url. Defaults to None.
            headers (Optional[Dict[str, str]]): Just headers, with which the request is sent. Defaults to None.
            wait (bool): wait result. Defaults to False.
            wrap (bool): need system wrap with operationId. Defaults to True.
            with_result (bool): return result. Defaults to False.
        """  # noqa: E501
        pass

    async def async_msg(self, data: Union[str, Dict], url: Optional[str] = None, headers: Optional[Dict[str, str]] = None, wait: bool = False, wrap: bool = True, with_result: bool = False):  # noqa: ANN201, E501
        """Sens http msg to system or any url

        Args:
            data (Union[str, Dict]): msg data
            url (Optional[str]): url to which the request is sent. If not specified, the default url is used (system url, but it can be overridden in the startup cfg, the parameter is msg_url). system url. Defaults to None.
            headers (Optional[Dict[str, str]]): Just headers, with which the request is sent. Defaults to None.
            wait (bool): wait result. Defaults to False.
            wrap (bool): need system wrap with operationId. Defaults to True.
            with_result (bool): return result. Defaults to False.
        """  # noqa: E501
        pass

    def email_send(self, message: str, subject: Optional[str] = None, type: str = "gmail") -> None:  # noqa: E501
        """Send an email

        Args:
            message (str): text message
            subject (Optional[str], optional): message subject, if subject is None used default. Defaults to None.
            type (str, optional): message type, only gmail work now. Defaults to "gmail".
        """  # noqa: E501
        pass

    def metadata(self, df_name: str) -> Optional[Union[Dict[str, Any], List[Dict[str, Any]]]]:  # noqa: E501
        """Gets metadata by df_name (if it saved with collection)

        Args:
            df_name (str): df name

        Returns:
            Optional[Union[Dict[str, Any], List[Dict[str, Any]]]]: metadata if exists (list if many), None otherwise
        """  # noqa: E501
        pass

    @property
    def scale_info(self) -> Tuple[int, int]:
        """Gets scale info: `index` and `index count`. `index count` - how many `apps` run it, `index` in [0, `index count`)

        Returns:
            Tuple[int, int]: `index` and `index count`
        """  # noqa: E501
        pass

    def get_scale_part(self, df: pd.DataFrame) -> pd.DataFrame:
        """Gets scale part of df (`index` and `index count` used for that) - all apps app get different data

        Args:
            df (pd.DataFrame): df to scale

        Returns:
            pd.DataFrame: scale part of df
        """  # noqa: E501
        pass

    @property
    def operation_id(self) -> str:
        """Operation identifier

        Returns:
            str: operation_id
        """
        pass

    def has_object(self, path: str) -> bool:
        """Checks whether there is an asset (OBJ) at the path

        Args:
            path (str): Path to object (asset)

        Returns:
            bool: exist or not
        """
        pass

    def get_object(self, path: str) -> OBJ:
        """Composes a path to an asset (OBJ) and returns it as :class:`OBJ`

        If the asset (object) does not exist, the function
        fails with an error.

        Args:
            path (str): Path to object (asset)

        Returns:
            OBJ: An OBJ wrap around a path
        """
        pass


    def as_object(
        self,
        paths: Dict[str, str],
        path_prefix: Optional[str] = None,
        *,
        dir: Optional[str] = None,
        allow_update_dir: bool = True
    ) -> OBJ:
        """Creates an asset (OBJ) by copying paths to specific directory and creating :class:`OBJ` by this dir.

        Assets (:class:`OBJ`) are simply a path within a directory accessible from within container
        by a certain user. This function creates a new asset by copying specified files into separate directory
        and creating an asset pointing to whole folder.

        Args:
            paths (Dict[str, str]): Path to an actual object -> subpath in asset directory
            path_prefix (Optional[str], optional): prefix for `paths` key if not None
            dir (Optional[str], optional): target directory name. If not set, it is generated. Defaults to None
            allow_update_dir (bool): raise exception if `dir` already exist and it set to False

        Returns:
            OBJ: OBJ with created directory captured
        """ # noqa: E501
        pass


def to_binary(smth: Any) -> bytes:  # noqa: ANN401
    """Converts object to binary
<<<<<<< HEAD
=======

>>>>>>> f9049b55f1efeadd27e32843d9c0d8c4431a7405
    Args:
        smth (Any): object to convert
    """
    return pickle.dumps(smth)


def from_binary(smth: bytes) -> Any:
    """Converts binary to object
<<<<<<< HEAD
=======

>>>>>>> f9049b55f1efeadd27e32843d9c0d8c4431a7405
    Args:
        smth (bytes): binary to convert
    """
    return pickle.loads(smth)


def load(url: str, path: str, path_prefix: str = APP_DIR) -> None:
    """Get request url and save result to path

    Args:
        url (str): url to load
        path (str): relative path - starting from the `path_prefix` (apps directory in app (`APP_DIR`) by default)
        path_prefix (str, optional): prefix for the path. Defaults to APP_DIR.
    """  # noqa: E501
    pass


class S3Helper:
    """
    Ready-made auxiliary wrapper for interacting with custom s3
    """

    def __init__(self, client: Any, s3_bucket: str) -> None:  # noqa: ANN401
        self.client = client
        self.__bucket = s3_bucket

    @staticmethod
    def create_by_cfg(cfg: Dict[str, Any], **kwargs) -> 'S3Helper':
        s3_client = boto3.client(
            's3',
            region_name=cfg.get('aws_region'),
            aws_access_key_id=cfg['aws_access_key_id'],
            aws_secret_access_key=cfg['aws_secret_access_key'],
            endpoint_url=cfg.get('endpoint_url'),
            **kwargs
        )
        return S3Helper(s3_client, cfg['s3_bucket'])

    def get_object(self, key: str, bucket: Optional[str] = None) -> Optional[StreamingBody]:  # noqa: E501
        """Uses :code:`get_object` from client by `bucket` and `key`

        Args:
            key (str): object storage key
            bucket (Optional[str], optional): object storage bucket, if set - override default. Defaults to None.

        Returns:
            Optional[StreamingBody]: get_object response body
        """  # noqa: E501
        pass

    def get_df(self, key: str, bucket: Optional[str] = None) -> pd.DataFrame:
        """Uses :code:`get_object` result and cast it to data frame

        Args:
            key (str): object storage key
            bucket (Optional[str], optional): object storage bucket, if set - override default. Defaults to None.

        Returns:
            pd.DataFrame: result in df view
        """  # noqa: E501
        pass

    def save_object(self, body: Any, key: str, bucket: Optional[str] = None) -> None:  # noqa: ANN401
        """Uses :code:`put_object` from client by `bucket`, `key` and `body`

        Args:
            body (Any): saved data
            key (str): object storage key
            bucket (Optional[str], optional): object storage bucket, if set - override default. Defaults to None.
        """  # noqa: E501
        pass

    def save_df(self, df: pd.DataFrame, key: str, bucket: Optional[str] = None) -> None:
        """Uses :code:`save_object` to save df by `bucket` and `key`

        Args:
            df (pd.DataFrame): df to save
            key (str): object storage key
            bucket (Optional[str], optional): object storage bucket, if set - override default. Defaults to None.
        """  # noqa: E501
        pass

    def delete_object(self, key: str, bucket: Optional[str] = None) -> None:
        """Deletes object by `bucket` and `key`

        Args:
            key (str): object storage key
            bucket (Optional[str], optional): object storage bucket, if set - override default. Defaults to None.
        """  # noqa: E501
        pass


class SmtpSender:
    """
    Ready-made auxiliary wrapper for interacting with SMTP
    Args:
        login (str): login
        password (str): password
        smtp_server (str, optional): smtp server. Defaults to "smtp.gmail.com".
        smtp_port (int, optional): smtp port. Defaults to 465.
    """

    def __init__(self, login: str, password: str, smtp_server: str = "smtp.gmail.com", smtp_port: int = 465) -> None:  # noqa: E501
        self.server = smtp_server
        self.port = smtp_port
        self.login = login
        self.password = password

    def send(self, receivers: list[str], subject: str, message: str) -> None:
        """Sends an email

        Args:
            receivers (list[str]): list of emails
            subject (str): message subject
            message (str): message text
        """
        pass


_Tensor = TypeVar('_Tensor', bound='torch.Tensor')
_kshort = 0b111  # last two-byte char encodes <= 7 bits
_kexclude_idx = {chr(0): 0, chr(10): 1, chr(
    13): 2, chr(34): 3, chr(38): 4, chr(92): 5}
_idx_exclude = {0: chr(0), 1: chr(10), 2: chr(
    13), 3: chr(34), 4: chr(38), 5: chr(92)}


def _base_encode(data: bytes) -> str:
    idx = (bit := 0)

    def get7(length):
        """get 7 bits from data"""
        nonlocal idx, bit, data
        if idx >= length:
            return False, 0

        # AND mask to get the first 7 bits
        f_ = (((0b11111110 % 0x100000000) >> bit) & data[idx]) << bit
        f_ = f_ >> 1
        bit += 7
        if bit < 8:
            return True, f_
        bit -= 8
        idx += 1
        if idx >= length:
            return True, f_
        secondPart = (((0xFF00 % 0x100000000) >> bit) & data[idx]) & 0xFF
        secondPart = secondPart >> (8 - bit)
        return True, f_ | secondPart

    _out = bytearray()
    while True:
        rbits, bits = get7(len(data))
        if not rbits:
            break
        if bits in _kexclude_idx:
            illegalIndex = _kexclude_idx[bits]
        else:
            _out.append(bits)
            continue
        retNext, nextBits = get7(len(data))
        b1 = 0b11000010
        b2 = 0b10000000
        if not retNext:
            b1 |= (0b111 & _kshort) << 2
            nextBits = bits
        else:
            b1 |= (0b111 & illegalIndex) << 2
        firstBit = 1 if (nextBits & 0b01000000) > 0 else 0
        b1 |= firstBit
        b2 |= nextBits & 0b00111111
        _out += [b1, b2]
    return ''.join([chr(x) for x in _out])


def _base_decode(encoded_data: str) -> bytes:
    encoded_data = [ord(x) for x in encoded_data]
    decoded = []
    curByte = bitOfByte = 0

    def push7(byte):
        nonlocal curByte, bitOfByte, decoded
        byte <<= 1
        curByte |= (byte % 0x100000000) >> bitOfByte
        bitOfByte += 7
        if bitOfByte >= 8:
            decoded += [curByte]
            bitOfByte -= 8
            curByte = (byte << (7 - bitOfByte)) & 255
        return

    for i in range(len(encoded_data)):
        if encoded_data[i] > 127:
            illegalIndex = ((encoded_data[i] % 0x100000000) >> 8) & 7
            if illegalIndex != _kshort:
                push7(_idx_exclude[illegalIndex])
            push7(encoded_data[i] & 127)
        else:
            push7(encoded_data[i])
    return bytearray(decoded)


def _tensor_to_df(x: list[_Tensor] | _Tensor) -> pd.DataFrame:
    import io

    import torch  # not in requirements

    if not isinstance(x, list):
        x = [x]

    shapes = []
    data = []
    grads = []
    device = []
    for x_ in x:
        assert isinstance(x_, torch.Tensor), f"not a tensor: {type(x_)}"
        # in-memory serialization using torch.save
        # to save autograd information and tensor type
        # https://pytorch.org/docs/stable/torch.html#torch.save
        buff = io.BytesIO()
        shape = x_.shape
        shapes.append(shape)
        device.append(x_.device)
        x_ = x_.cpu()
        torch.save(x_, buff)
        buff.seek(0)
        data.append(_base_encode(buff.read()))

        buff.close()
        buff = io.BytesIO()
        torch.save(x_.grad, buff)
        buff.seek(0)

        grads.append(_base_encode(buff.read()))

    return pd.DataFrame(
        {
            "__shape__": shapes,
            "__tensor__": data,
            "__grad__": grads,
            "__device__": device,
        })


def _tensor_from_df(x: pd.DataFrame) -> list:
    import io

    import torch  # not in requirements

    _out = []
    for _, row in x.iterrows():
        shape = row["__shape__"]
        encoded = row["__tensor__"]
        encoded_grad = row["__grad__"]
        decoded = _base_decode(encoded)
        decoded_grad = _base_decode(encoded_grad)
        buff = io.BytesIO(decoded)
        buff.seek(0)
        _t = torch.load(buff).reshape(shape)

        if _t.requires_grad:
            buff.close()

            buff = io.BytesIO(decoded_grad)
            buff.seek(0)
            _t.grad = torch.load(buff).reshape(shape)
            buff.close()

        if row["__device__"] != "cpu" and torch.cuda.is_available():
            _t = _t.to(row["__device__"])

        _out.append(_t)

    return _out


def to_df(x: Any, force: bool = False) -> pd.DataFrame:
    """Creates a data frame from an arbitrary object
    - `torch.Tensor`: Tensor is serialized using torch.save and then encoded using base112. Autograd information is preserved.
    - `numpy`, `list`, `tuple`, `range`, `bytearray`: Data is serialized using pickle and stored as is in `data` column.
    - `set`, `frozenset`: Data is converted to list and stored as is in `data` column.
    - `dict`: Data is serialized using json and stored as is in `data` column.
    - `int`, float, complex, str, bytes, bool: Data is stored as is in `data` column.


    Args:
        x (Any): Object to convert to data frame
        force (bool, optional):
            If set, it will ignore the type of the object and serialize it using pickle.
            Defaults to False.

    Returns:
        pd.DataFrame: Data frame with a single column :code:`data`
    """ # noqa: E501
    if force:
        return pd.DataFrame({"data": [jsonpickle.encode(x)]})
    elif type(x).__name__ == "Tensor" or (isinstance(x, list) and len(x) > 0 and type(x[0]).__name__ == "Tensor"):
        return _tensor_to_df(x)
    elif isinstance(x, (np.ndarray, list, tuple, range, bytearray)):
        return pd.DataFrame({"data": x})
    elif isinstance(x, (set, frozenset)):
        return pd.DataFrame({"data": list(x)})
    elif isinstance(x, dict):
        return pd.DataFrame({"data": [json.dumps(x)]})
    else:   # int, float, complex, str, bytes, bool
        return pd.DataFrame({"data": [x]})


# TODO create same with pyspark
def from_df(x: pd.DataFrame, type_name: Optional[str] = None, force: bool = False) -> Any:  # noqa: ANN401, E501
    """Converts a data frame obtained by running :func:`to_df` back to an object

    Args:
        x (pd.DataFrame): Data frame to convert
        type_name (Optional[str], optional):
            Type of the object to convert to. If not specified, the type is inferred from the data frame.
            Defaults to None.
        force (bool, optional):
            If set, it will ignore the type of the object and deserialize it using pickle.
            Defaults to False.

    Returns:
        Any: Object of type :code:`type_name` or inferred type
    """
    if force:
        return jsonpickle.decode(x.data[0])
    elif type_name == 'ndarray':
        return x.data.values
    elif type_name == 'list':
        return x.data.values.tolist()
    elif type_name == 'tuple':
        return tuple(x.data.values.tolist())
    elif type_name == 'range':
        return x.data.values.tolist()
    elif type_name == 'Tensor' or ('__shape__' in x.columns and '__tensor__' in x.columns):
        # import torch  # not in requirements
        # return torch.from_numpy(x.values).float().to(torch.device('cpu'))   # can't work with gpu from inside yet  # noqa: E501
        return _tensor_from_df(x)
    elif type_name == 'set':
        return set(x.data.values.tolist())
    elif type_name == 'frozenset':
        return frozenset(x.data.values.tolist())
    elif type_name == 'dict':
        return json.loads(x.data[0])
    elif type_name == 'bytearray':
        return bytearray(x.data)
    else:   # int, float, complex, str, bytes, bool
        return x.data[0]



from dataclasses import dataclass
from enum import Enum


class ImageCompatStrategy(Enum):
    FULL_REF = 0
    ONLY_TAG = 1
    ONLY_NAME = 2


class SpaceCompatStrategy(Enum):
    REVERSE_ID = 0b0001
    IMAGE = 0b0010
    VERSION = 0b0100
    BRANCH = 0b1000

@dataclass
class CompatabilityStrategy:
    image: int = ImageCompatStrategy.ONLY_NAME
    space: int = SpaceCompatStrategy.REVERSE_ID
    none_is_always_compatible: bool = False



def compare_images(
    image_ref: str,
    other_image_ref: str,
    strategy: CompatabilityStrategy
) -> bool:
    if image_ref is None or other_image_ref is None:
        return strategy.none_is_always_compatible
    if strategy.image == ImageCompatStrategy.FULL_REF:
        return image_ref == other_image_ref
    elif strategy.image == ImageCompatStrategy.ONLY_TAG:
        return image_ref.split(":")[1] == other_image_ref.split(":")[1]
    elif strategy.image == ImageCompatStrategy.ONLY_NAME:
        return image_ref.split(":")[0] == other_image_ref.split(":")[0]
    else:
        raise ValueError("Invalid strategy")



from .flow_function import FlowFunction
from .exceptions import InterpretationError
from .actions import Action
from .preferences import UserPreferences, VerbosityLevel
from .callbacks import (
    SpaceCallbackBody,
    SpaceCallbackResult
)


from .overrides import (
    Override,
    AssetOverride,
    DocumentOverride,
    CollectionOverride
)

from .injections import CoreInjectable, SpaceInjectable, BaseInjectable
from .registry.core_entry import CoreRegistryEntry
from .manifest import Manifest, ManifestUpdateEntry, Secret, Secrets
from .dependency import Dependency, Integration
from .installers import (
    CompatabilityStrategy,
    ImageCompatStrategy,
    SpaceCompatStrategy,
    ImageOptions,
    ImageDependency,
    SpaceDependency,
    SpaceDependencyOptions
)
from .python_string import PythonString
from .argument import ArgumentLink
from .collection import Collection

from .nodes import (
    BaseNode,
    AssetNode,
    TreeNode,
    OperationNode,
    CollectionNode,
    DocumentNode
)

from .state import (
    CoreInterpreterState,
    CoreParams,
    SpaceAuxParams,
    SpaceInterpreterState,
)


from .results import (
    Result,
    BaseResult,
    CoreLocalDFResult,
    CoreResult,
    CoreResultPayload,
    SpaceCollectionResult
)



from .endpoint import MetaEndpoint
from .task import BaseTask
from .task import (
    PromisedStage,
    PromisedTask,
    CoreTask,
    CoreTaskState,
    CoreTaskStage,
    SpaceTask,
    SpaceTaskStage
)

from .annotations import ConfigArgument
from .in_app_core_info import InjectedAppInfo
from .annotations import ConfigArgument
from ._model import _Model


from .mappers import (
    AddPositiveCondition,
    AddNegativeCondition
)



import os
import shutil
from typing import Annotated

import rich
import typer

from malevich._utility import package_manager


def new(
    app_name: Annotated[str, typer.Argument(...)],
    base_path: Annotated[
        str, typer.Option("--path", "-p", show_default=False)
    ] = os.getcwd(),
    force: Annotated[bool, typer.Option("--force", "-f", show_default=False)] = False,
) -> None:
    dockerfile_ = os.path.join(
        package_manager.get_malevich_path(),
        "_templates",
        "Dockerfile.app"
    )

    if os.path.exists(os.path.join(base_path, app_name)) and not force:
        rich.print(
            f"[red]App {app_name} already exists. Use -f to overwrite[/red]"
        )
    elif os.path.exists(os.path.join(base_path, app_name)) and force:
        shutil.rmtree(
            os.path.join(base_path, app_name)
        )

    os.makedirs(
        os.path.join(
            base_path,
            app_name
        ),
        exist_ok=True
    )

    os.makedirs(
        os.path.join(
            base_path,
            app_name,
            "apps"
        ),
        exist_ok=True
    )

    shutil.copyfile(
        dockerfile_,
        os.path.join(
            base_path,
            app_name,
            "Dockerfile"
        )
    )

    with open(
        os.path.join(
            base_path,
            app_name,
            'apps',
            "processors.py"
        ),
        "w"
    ) as f:
        f.write(
            open(
                os.path.join(
                    package_manager.get_malevich_path(),
                    "_templates",
                    "processor.py.txt"
                )
            ).read()
        )

    with open(
        os.path.join(
            base_path,
            app_name,
            "flow.py"
        ),
        "w"
    ) as f:
        f.write(
            open(
                os.path.join(
                    package_manager.get_malevich_path(),
                    "_templates",
                    "flow.py.txt"
                )
            ).read().format(app=app_name)
        )

    with open(
        os.path.join(
            base_path,
            app_name,
            "requirements.txt"
        ),
        "w"
    ) as f:
        f.write("# Add your dependencies here\n")

    with open(
        os.path.join(
            base_path,
            app_name,
            "README.md"
        ),
        "w"
    ) as f:
        f.write(
            open(
                os.path.join(
                    package_manager.get_malevich_path(),
                    "_templates",
                    "README.app.md"
                )
            ).read().format(app=app_name)
        )



import json
import warnings
from functools import cache
from typing import Optional, TypeVar

import malevich_coretools as core
import pandas as pd
from pydantic import BaseModel

from malevich.constants import DEFAULT_CORE_HOST
from malevich.models import Collection

from ..base import BaseResult

DocumentModelType = TypeVar('DocumentModelType', bound=BaseModel)

class CoreResultPayload:
    """An actual information that is saved as result

    A payload can be either asset or collection. If it is an asset,
    it can be a single file or a composite asset (a directory with
    multiple files). It represents one of results of a single operation.

    Payload follows the following convention:

    - If :meth:`is_asset` is True, then :attr:`data` is :class:`bytes` object
    - If :meth:`is_collection` is True, then :attr:`data` is :class:`DataFrame` object
    - If :meth:`is_composite_asset` is True, then :attr:`data` is a dict of :class:`bytes` objects

    """  # noqa: E501

    def __init__(
        self,
        data: pd.DataFrame | dict | list[bytes],
        is_asset: bool = False,
        is_composite_asset: bool = False,
        is_document: bool = False,
        is_collection: bool = False,
        paths: Optional[list[str]] = None,
    ) -> None:
        assert not (is_asset and is_collection), \
            "Result can't be both asset and collection"
        if is_composite_asset:
            assert is_asset, \
                "Composite asset must be an asset"

        self._data = data
        self._is_asset = is_asset or is_composite_asset
        self._is_composite_asset = is_composite_asset
        self._is_collection = is_collection
        self._paths = paths or []
        self.is_document = is_document

    def is_asset(self) -> bool:
        """Checks if the result is an asset

        Returns:
            bool: True if the result is an asset
        """
        return self._is_asset

    def is_composite_asset(self) -> bool:
        """Checks if the result is a composite asset

        Returns:
            bool: True if the result is a composite asset
        """
        return self._is_composite_asset

    def is_collection(self) -> bool:
        """Checks if the result is a collection

        Returns:
            bool: True if the result is a collection
        """
        return self._is_collection

    def count_objects(self) -> int:
        if self._is_composite_asset:
            return len(self._paths)
        if self._is_asset:
            return 1
        if self._is_collection:
            return len(self._data.index)

    @property
    def data(self) -> pd.DataFrame | dict[str, bytes] | bytes | core.ResultCollection:
        if self._is_asset:
            if self._is_composite_asset:
                return {
                    path: self._data[i]
                    for i, path in enumerate(self._paths)
                }
            else:
                return self._data[0]
        else:
            return self._data

    def get_path(self) -> str:
        assert len(self._paths) == 1, (
            '`get_path` only works with exactly 1 path, '
            f'but there are {len(self._paths)}'
        )

        return self._paths[0]

    def __str__(self) -> str:
        return (
            f"CoreResultPayload(is_asset={self._is_asset}, "
            f"is_collection={self._is_collection} "
            f"is_composite_asset={self._is_composite_asset} "
            f"paths={self._paths})"
        )

    def __repr__(self) -> str:
        return (
            f"CoreResultPayload(is_asset={self._is_asset}, "
            f"is_collection={self._is_collection} "
            f"is_composite_asset={self._is_composite_asset} "
            f"paths={self._paths})"
        )


class CoreResult(BaseResult[CoreResultPayload]):
    """A representation of a result of execution of a flow on Malevich Core

    Represents a single returned result, which can contain a collection,
    a list of collections, a single asset or a list of assets.

    To be precise, for the function:

    .. code-block:: python

            from malevich import flow, collection, CoreInterpreter
            from malevich.etl import process
            from malevich.utility import merge

            @flow()
            def data_process():
                data = collection('my_fancy_smth', df=pd.DataFrame({'a': [1, 2, 3]}))
                processed = process(data)
                output = merge(
                    processed,
                    data,
                    config={'how': 'inner', 'on': 'index'}
                )
                return output, processed

            task = data_process()
            task.interpret(CoreInterpreter())
            # list of CoreResult objects
            results = task()
            # single CoreResult object
            output = results[0]
            # list of CoreResultPayload objects
            print(results[0].get())

    The `output` variable will be a single CoreResult object, which contains
    a single CoreResultPayload object, which contains a single data frame
    (the result of the merge operation). The `processed` variable will be
    a single CoreResult object, which contains a single CoreResultPayload
    object, which contains a single collection (the result of the process
    operation).

    **Note**: The class is not intended to be constructed by user.
    """

    @staticmethod
    def is_asset(data: pd.DataFrame) -> bool:
        return len(data.columns) > 0 and data.columns[0] == "path"

    @staticmethod
    def extract_path_to_asset(path: str, user: str) -> str:
        return path.split(f"/mnt_obj/{user}/")[-1]

    def __init__(
        self,
        core_group_name: str,
        core_operation_id: str,
        core_run_id: str,
        conn_url: str,
        auth: core.AUTH,
    ) -> None:
        self.core_group_name = core_group_name
        self._conn_url = conn_url
        self._auth = auth
        self.core_operation_id = core_operation_id
        self.core_run_id = core_run_id

    @property
    def num_elements(self) -> int:
        """The number of elements (assets/collections) in the result"""
        return super().num_elements

    def __str__(self) -> str:
        return f'CoreResult(core_operation_id="{self.core_operation_id}")'

    def __repr__(self) -> str:
        return f'CoreResult(core_operation_id="{self.core_operation_id}")'

    @cache
    def get(self) -> list[CoreResultPayload]:
        """Retrieves results from the Core

        Returns a list of CoreResultPayload objects, each of which
        contains an output from a single operation. The output might
        be a collection, a list of collection, a single file (asset)
        or a list of files (composite asset).

        If you wish to get more specific results, use the following methods:

        - :meth:`get_df` to get a single data frame (if result is a collection)
        - :meth:`get_dfs` to get a list of data frames (if result is a collection)
        - :meth:`get_binary` to get a single binary (if result is an asset)
        - :meth:`get_binary_dir` to get a dict of binaries (if result is a composite asset)

        Returns:
            list[CoreResultPayload]: The list of results

        """  # noqa: E501
        collections = [
            x for x in core.get_collections_by_group_name(
                self.core_group_name,
                operation_id=self.core_operation_id,
                run_id=self.core_run_id,
                auth=self._auth,
                conn_url=self._conn_url
            ).data
        ]

        results = []
        for col in collections:
            if '#' in col.id:
                results.append(CoreResultPayload(
                    data=json.loads(col.docs[0]),
                    is_document=True,
                ))
                continue

            result = pd.DataFrame([json.loads(j.data) for j in col.docs])
            if CoreResult.is_asset(result):
# if asset
                # NOTE: Path now returned without /mnt_obj/<user>
                # prefix, but I left the code as it was
                # before, just in case

                obj_path = CoreResult.extract_path_to_asset(
                    result.path.iloc[0],
                    user=self._auth[0],
                )
                # ================================================
                try:
                    objects_ = core.get_collection_objects(
                        obj_path,
                        auth=self._auth,
                        conn_url=self._conn_url,
                        # NOTE: Maybe allow getting with recursive
                        recursive=True
                    )
# # if one file
                    if len(objects_.files) == 1:
                        object_ = core.get_collection_object(
                            obj_path + "/" + objects_.files[0],
                            auth=self._auth,
                            conn_url=self._conn_url,
                        )
                        results.append(CoreResultPayload(
                            data=[object_],
                            is_asset=True,
                            paths=[obj_path + "/" + objects_.files[0]]
                        ))
                    else:
# # if multiple files
                        paths = []
                        obj_bytes = []
                        for file_ in objects_.files:
                            object_ = core.get_collection_object(
                                obj_path.rstrip("/") + "/" + file_.lstrip("/"),
                                auth=self._auth,
                                conn_url=self._conn_url,
                            )
                            paths.append(file_)
                            obj_bytes.append(object_)
                        results.append(CoreResultPayload(
                            data=obj_bytes,
                            is_asset=True,
                            is_composite_asset=True,
                            paths=paths,
                        ))
# if failed, try without recursive
                except Exception as _:
                    try:
                        object_ = core.get_collection_object(
                            obj_path,
                            auth=self._auth,
                            conn_url=self._conn_url,
                        )
                        results.append(CoreResultPayload(
                            data=[object_],
                            is_asset=True,
                            paths=[obj_path],
                        ))
# totall failure = keep it as a collection
                    except Exception as _:
                        results.append(CoreResultPayload(
                            data=result,
                            is_collection=True,
                        ))
# if collection
            else:
                results.append(CoreResultPayload(
                    data=result,
                    is_collection=True,
                ))

        return results

    @cache
    def get_df(self) -> pd.DataFrame:
        """Converts the result to a single data frame

        It only works if the result is a single collection,
        otherwise it raises an error. Check the number of
        elements using :attr:`num_elements` property and use only if
        it is equal to 1

        Returns:
            DataFrame: The result as a data frame

        Raises:
            NotImplementedError: If the result is not a single collection
        """
        if result := self.get():
            if len(result) == 1:
                if result[0].is_collection():
                    return result[0].data
                elif result[0].is_asset():
                    return NotImplementedError(
                        "Cannot return a single DataFrame from an asset. "
                        "Please use `get_binary` or `get_binary_dir` instead"
                    )
            else:
                raise NotImplementedError(
                    "Cannot return a single DataFrame from multiple results. "
                    "Please use `get_dfs` instead and check the number of results "
                    "using `num_elements` property"
                )
        else:
            warnings.warn(f"No results found for {self.core_group_name}")
            return pd.DataFrame()

    @cache
    def get_dfs(self) -> list[pd.DataFrame]:
        """Converts the result to a list of data frames

        If some of the results are assets, they are ignored
        in the output.

        Returns:
            list[:class:`DataFrame`]: The result as a list of data frames
        """
        if result := self.get():
            results_ = []
            for res in result:
                if res.is_collection():
                    results_.append(res.data)
                elif res.is_asset():
                    continue
            return results_
        else:
            return []

    @cache
    def get_binary(self) -> bytes:
        """Retrieves asset binary data, if the result is file asset

        It only works if the result is a single file returned
        as asset. Otherwise it raises an error. Check the number of
        elements using :attr:`num_elements` property and use only if
        it is equal to 1

        Returns:
            bytes: The binary data of the asset

        Raises:
            NotImplementedError: If the result is not a single asset
        """
        if result := self.get():
            if len(result) == 1:
                if result[0].is_collection():
                    raise NotImplementedError(
                        "Cannot return a single binary from a collection. "
                        "Please use `get_df` or `get_dfs` instead"
                    )
                elif result[0].is_asset():
                    return result[0].data
            else:
                raise NotImplementedError(
                    "Cannot return a single binary from multiple results. "
                    "Please use `get_binary_dir` instead and "
                    "check the number of results "
                    "using `num_elements` property"
                )
        else:
            warnings.warn(f"No results found for {self.core_group_name}")
            return b""

    @cache
    def get_binary_dir(self) -> dict[str, bytes]:
        """Retrieves files from assets

        Non-assets are ignored in the output

        Returns:
            dict[str, bytes]: Dict of file names and their binary data
        """
        if result := self.get():
            results_ = {}
            for res in result:
                if res.is_collection():
                    raise NotImplementedError(
                        "Cannot return a binary directory from a collection. "
                        "Please use `get_df` or `get_dfs` instead"
                    )
                elif res.is_asset():
                    if res.is_composite_asset():
                        results_.update(res.data)
                    else:
                        results_[res.get_path()] = res.data
            return results_
        else:
            warnings.warn(f"No results found for {self.core_group_name}")
            return {}

    @cache
    def get_document(
        self,
        model: type[DocumentModelType] | None = None
    ) -> dict | DocumentModelType:
        """Retrieves the document of the result

        Returns:
            dict: The document of the result
        """
        if result := self.get():
            # if result[index].is_document:
            #     return result[index].data if model is None else model(
            #         **result[index].data
            #     )
            # else:
            #     what = "asset" if result[index].is_asset() else "collection"
            #     raise NotImplementedError(
            #         "Cannot return a document from a non-document result. "
            #         f"Result at index {index} is {what}"
            #     )
            data_ = result[0].data.to_dict(orient='records')[0]
            return model(**data_) if model else data_

    @cache
    def get_documents(
        self,
        model: type[DocumentModelType] | None = None
    ) -> list[dict] | list[DocumentModelType]:
        if result := self.get():
            # if result[index].is_document:
            #     return result[index].data if model is None else model(
            #         **result[index].data
            #     )
            # else:
            #     what = "asset" if result[index].is_asset() else "collection"
            #     raise NotImplementedError(
            #         "Cannot return a document from a non-document result. "
            #         f"Result at index {index} is {what}"
            #     )
            data_ = result[0].data.to_dict(orient='records')
            if model:
                return [model(**d) for d in data_]
            return data_



class CoreLocalDFResult(BaseResult[pd.DataFrame]):
    """A specification of :func:`collection` as a result

    This class is used to wrap a collection object as result.
    It simply stores collection object as a DataFrame and
    returns it when :meth:`get` method is called.
    """

    def __init__(
        self,
        coll: Collection,
        auth: core.AUTH,
        conn_url: str = DEFAULT_CORE_HOST,
    ) -> None:
        self._coll = coll
        self._auth = auth
        self._conn_url = conn_url

    def num_elements(self) -> int:
        """The number of elements (assets/collections) in the result"""
        return 1

    def get(self) -> pd.DataFrame | None:
        """Simply extracts saved data frame

        Returns:
            :class:`DataFrame`: Saved data frame
        """
        if self._coll.collection_data is not None:
            return self._coll.collection_data

        # NOTE: Maybe it is better to try
        # to fetch collection from Core?

        # try:
        #     ids = core.get_collections_by_name(
        #         self._coll.magic(),
        #         auth=self._auth,
        #         conn_url=self._conn_url,
        #     ).ownIds

        #     if len(ids) == 1:
        #         return core.get_collection_to_df(
        #             ids[0],
        #             auth=self._auth,
        #             conn_url=self._conn_url,
        #         )
        #     else:
        #         return [
        #             core.get_collection_to_df(
        #                 i,
        #                 auth=self._auth,
        #                 conn_url=self._conn_url,
        #             )
        #             for i in ids

        #         ]
        # except Exception as _:
        #     return None




from hashlib import sha256

from ..models.nodes.asset import AssetNode
from ..models.nodes.collection import CollectionNode
from ..models.nodes.document import DocumentNode
from ..models.nodes.operation import OperationNode
from ..models.nodes.tree import TreeNode
from ..types import FlowTree
from .tree import unwrap_tree


def get_tree_node_hash(tree_node: TreeNode | FlowTree) -> str:
    all_edges = set()
    tree = tree_node.tree if isinstance(tree_node, TreeNode) else tree_node
    for from_, to_, link_ in unwrap_tree(tree).tree:
        from_ = from_.owner
        to_ = to_.owner
        if isinstance(from_, OperationNode):
            fk = from_.processor_id
        elif isinstance(from_, CollectionNode):
            fk = from_.collection.collection_id
        elif isinstance(from_, AssetNode):
            fk = from_.name
        elif isinstance(from_, DocumentNode):
            fk = from_.magic() if from_.document is not None else from_.reverse_id

        if isinstance(to_, OperationNode):
            tk = to_.processor_id
        elif isinstance(to_, CollectionNode):
            tk = to_.collection.collection_id
        elif isinstance(to_, AssetNode):
            tk = to_.name
        elif isinstance(to_, DocumentNode):
            tk = to_.magic() if to_.document is not None else to_.reverse_id

        all_edges.add((fk, tk, link_.index))

    all_edges = sorted(all_edges, key=lambda x: x[2])
    return sha256(str(all_edges).encode()).hexdigest()



import concurrent.futures

import rich
from rich.progress import Progress, SpinnerColumn, TextColumn

from malevich._utility import FlowStub
from malevich.install import (
    ImageInstaller,
    Installer,
    SpaceInstaller,
)
from malevich.manifest import ManifestManager
from malevich.models.dependency import Integration

from .prefs import prefs as prefs


def _restore(installer: Installer, depedency: dict, progress: Progress, task) -> None:
    try:
        package_id = depedency["package_id"]
        parsed = installer.construct_dependency(depedency)

        installer.restore(parsed)
        progress.update(
            task,
            completed=True,
            description=f"[green]✔[/green] Package [b blue]{package_id}[/b blue]",
        )
    except Exception as e:
        if task:
            progress.update(
                task,
                completed=True,
                description=f"[red]✘[/red] Package [b blue]{package_id}[/b blue]",
            )
        return e, package_id


# @app.command(name="restore", help=restore_help["--help"])
def restore() -> None:
    manf = ManifestManager()
    with Progress(
        SpinnerColumn(), TextColumn("{task.description}")
    ) as progress, concurrent.futures.ThreadPoolExecutor() as executor:
        futures = []
        for record in manf.query("dependencies"):
            key = next(iter(record.keys()))
            dependency = record[key]
            installed_by = dependency["installer"]
            task = progress.add_task(
                f"Package [green]{dependency['package_id']}[/green] with "
                f"[yellow]{installed_by}[/yellow]",
                total=1
            )
            if installed_by == "image":
                image_installer = ImageInstaller()
                futures.append(
                    executor.submit(_restore, image_installer,
                                    dependency, progress, task)
                )
            elif installed_by == 'space':
                space_installer = SpaceInstaller()
                futures.append(
                    executor.submit(_restore, space_installer,
                                    dependency, progress, task)
                )

    for future in concurrent.futures.as_completed(futures):
        result = future.result()
        if result:
            progress.stop()
            rich.print(f"[red]Failed to restore package {result[1]}[/red]")
            rich.print(result[0])
            exit(-1)

    progress.stop()
    rich.print("\n\n[b]Restoring integrations:[/b]")
    for flow, integrations in manf.query('flows').items():
        FlowStub.sync_flows(flow, integrations=[
            Integration(**i) for i in integrations
        ])
        rich.print(
            f"[green]✔[/green] [green]{flow}[/green] ({len(integrations)} integrations)"
        )



import rich
from typer import Typer

from malevich.core_api import (
    delete_endpoint,
    get_endpoint,
    get_endpoints,
    get_run_endpoint,
    pause_endpoint,
    resume_endpoint,
)

from ..misc.make import wrap_command

endpoints_app = Typer(name='endpoints')


@endpoints_app.command(name='list')
@wrap_command(get_endpoints)
def list_endpoints(verbose: bool = False, **kwargs) -> None:
    endpoints = get_endpoints(**kwargs)
    if not verbose:
        rich.print(
            '\n'.join([f'{e.hash} {e.description}' for e in endpoints.data]))
    else:
        rich.print(endpoints.data)


@endpoints_app.command(name='get')
@wrap_command(get_endpoint)
def get(**kwargs) -> None:
    endpoint = get_endpoint(**kwargs)
    rich.print(endpoint)


@endpoints_app.command(name='pause')
@wrap_command(pause_endpoint)
def pause(**kwargs) -> None:
    pause_endpoint(**kwargs)
    rich.print('Endpoint paused')


@endpoints_app.command(name='resume')
@wrap_command(resume_endpoint)
def resume(**kwargs) -> None:
    resume_endpoint(**kwargs)
    rich.print('Endpoint resumed')


@endpoints_app.command(name='delete')
@wrap_command(delete_endpoint)
def delete(**kwargs) -> None:
    delete_endpoint(**kwargs)
    rich.print('Endpoint deleted')


@endpoints_app.command(name='run-options')
@wrap_command(get_run_endpoint)
def run_options(**kwargs) -> None:
    run = get_run_endpoint(**kwargs)
    rich.print(run)



from sqlalchemy import Column, Integer, Sequence, String, create_engine
from .base import Base

class CachedCredentials(Base):
    __tablename__ = 'tb_creds'
    
    id = Column(Integer, Sequence('tb_creds_id'), primary_key=True)
    email = Column(String(255))
    api_url = Column(String(255))
    host = Column(String(255))
    org_id = Column(String(255), nullable=True)
    core_username = Column(String(255))
    core_password = Column(String(255))
    



import json
from typing import Iterable, Optional

from pydantic import ConfigDict

from malevich._autoflow.tracer import ExecutionTree, traced
from malevich.models import (
    ArgumentLink,
    PythonString,
)

from .base import BaseNode


class TreeNode(BaseNode):
    tree: ExecutionTree[traced[BaseNode], ArgumentLink[BaseNode]] | None = None
    results: Iterable[traced[BaseNode]] | traced[BaseNode] | None = None

    # Space Component
    reverse_id: PythonString
    name: str
    description: str = "Wonderful Flow!"
    version_uid: str | None = None
    branch_uid: str | None = None
    # ================ #

    underlying_node: Optional['BaseNode'] = None
    integrated: bool = False
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def get_senstivite_fields(self) -> dict[str, str]:
        in_nodes_fields = {}
        for u, v, _ in self.tree.tree:
            if u.owner.uuid not in in_nodes_fields:
                in_nodes_fields[u.owner.uuid] = u.owner.get_senstivite_fields()
            if v.owner.uuid not in in_nodes_fields:
                in_nodes_fields[v.owner.uuid] = v.owner.get_senstivite_fields()
        results_fields = {}
        if self.results is not None:
            if isinstance(self.results, Iterable):
                for node in self.results:
                    results_fields[node.owner.uuid] = node.owner.get_senstivite_fields()
            else:
                results_fields[self.results.owner.uuid] = (
                    self.results.owner.get_senstivite_fields()
                )
        return {
            "in_nodes_fields": json.dumps(in_nodes_fields),
            "results_fields": json.dumps(results_fields),
        }

    def set_sensitive_fields(self, values: dict[str, str]) -> None:
        in_nodes_fields = values.get("in_nodes_fields")
        if in_nodes_fields is not None:
            in_nodes_fields = json.loads(in_nodes_fields)
            for u, v, _ in self.tree.tree:
                u.owner.set_sensitive_fields(in_nodes_fields[u.owner.uuid])
                v.owner.set_sensitive_fields(in_nodes_fields[v.owner.uuid])
        results_fields = values.get("results_fields")
        if results_fields is not None:
            results_fields = json.loads(results_fields)
            if isinstance(self.results, Iterable):
                for node in self.results:
                    node.owner.set_sensitive_fields(results_fields[node.owner.uuid])
            else:
                self.results.owner.set_sensitive_fields(results_fields[self.results.owner.uuid])

    def __hash__(self) -> int:
        return super().__hash__()


    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, TreeNode):
            return False

        eq_ = self.uuid == __value.uuid
        if self.underlying_node or __value.underlying_node:
            eq_ &= self.underlying_node == __value.underlying_node
        return eq_



import os

import pandas as pd
from pydantic import BaseModel, ConfigDict


class Override(BaseModel): ...

class AssetOverride(Override):
    path: str
    file: os.PathLike | None
    folder: os.PathLike | None
    files: list[os.PathLike] | None

class CollectionOverride(Override):
    model_config = ConfigDict(
        arbitrary_types_allowed=True
    )

    data: pd.DataFrame

class DocumentOverride(Override):
    data: BaseModel




import os

from malevich_space.constants import *  # noqa: F403

from .path import MALEVICH_CACHE, MALEVICH_HOME, Paths  # noqa: F401

DEFAULT_CORE_HOST = "https://core.malevich.ai/"

APP_HELP = """
Malevich is a co
"""

USE_HELP = """
Install apps to get access to the processors they provide
"""

USE_IMAGE_HELP = """
Install apps using published Docker images. The information about
the functionality provided by them will be pulled and parsed accordingly.
"""

USE_SPACE_HELP = """
Install components from Malevich Space repository.
"""

IMAGE_BASE = "public.ecr.aws/o1z1g3t0/{app}:latest"

CORE_INTERPRETER_IN_APP_INFO_KEY = '__core__'

reserved_config_fields = [
    ('alias', 'str'),
]

TEST_DIR =  os.path.expanduser(
    os.getenv("MALEVICH_TEST_DIR", Paths.home('testing', create_dir=True))
)




import logging
import re
from typing import Annotated

import rich
import typer
from rich.progress import Progress, SpinnerColumn, TextColumn

from malevich._deploy import Space
from malevich._utility import parse_kv_args
from malevich.install.flow import FlowInstaller
from malevich.models.dependency import Integration

from .prefs import prefs as prefs
from .use import _install_from_image, _install_from_space

logging.getLogger("gql.transport.requests").setLevel(logging.ERROR)


__With_Args_Help = (
    "Arguments to pass to the installer. "
    "Must be in the format [b]key1=value1,key2=value2[/b] "
    '(" can be used to escape spaces)'
    '\n\n\n'
    'For a particular installer, use [i b] malevich use <installer> --help[/i b] to'
    ' see the list of supported arguments'
)

def auto_use(
    package_names: Annotated[list[str], typer.Argument(...)],
    using: Annotated[str, typer.Option(help="Installer to use")] = "space",
    with_args: Annotated[str, typer.Option(help=__With_Args_Help)] = "",
) -> None:
    rich.print(
        "\n\nAttempting automatic installation of components [b blue]"
        f"{'[white], [/white]'.join(package_names)}[/b blue]\n"
    )
    args = parse_kv_args(with_args)
    if args:
        rich.print(
            "\n".join(
                [
                    f"\t[green_yellow]{key}[/green_yellow]: [white]{value}[/white]"
                    for key, value in args.items()
                ]
            )
        )
        rich.print("\n\n")

    with Progress(SpinnerColumn(), TextColumn("{task.description}")) as progress:
        if using == "image":
            for package_name in package_names:
                task = progress.add_task(
                    f"Attempting to install [b blue]{package_name}[/b blue] with"
                    " [i yellow]image[/i yellow] installer",
                    total=1,
                )
                try:
                    if "package_name" not in args:
                        args["package_name"] = package_name
                    _install_from_image(**args)
                    args.pop("package_name")
                except Exception as err:
                    progress.update(
                        task,
                        description="[red]✘[/red] Failed with "
                        "[yellow]image[/yellow] installer "
                        f"[blue]({package_name})[/blue]",
                        completed=1,
                    )
                    progress.stop()
                    rich.print("\n\n[red]Installation failled[/red]")
                    rich.print(err)
                    raise err
                    exit(-1)
                else:
                    progress.update(
                        task,
                        description="[green]✔[/green] Success with "
                        f"[yellow]image[/yellow][blue] ({package_name})[/blue]",
                        completed=1,
                    )
            progress.stop()
            rich.print(
                f"\n\nInstallation of [blue]{'[white], [/white]'.join(package_names)}"
                "[/blue] with [yellow]image[/yellow] installer was "
                "[green]successful[/green]"
            )

        elif using == "space":
            for package_name in package_names:
                task = progress.add_task(
                    f"Attempting to install [b blue]{package_name}[/b blue] with"
                    " [i yellow]space[/i yellow] installer",
                    total=1,
                )
                try:
                    if "package_name" not in args:
                        args["package_name"] = package_name
                    args["reverse_id"] = package_name
                    _install_from_space(**args)
                    args.pop("package_name")
                except Exception as err:
                    progress.update(
                        task,
                        description="[red]✘[/red] Failed with "
                        "[yellow]space[/yellow] installer "
                        f"[blue]({package_name})[/blue]",
                        completed=1,
                    )
                    progress.stop()
                    rich.print("\n\n[red]Installation failled[/red]")
                    raise err
                else:
                    progress.update(
                        task,
                        description="[green]✔[/green] Success with "
                        f"[yellow]space[/yellow][blue] ({package_name})[/blue]",
                        completed=1,
                    )
            progress.stop()
            rich.print(
                f"\n\nInstallation of [blue]{'[white], [/white]'.join(package_names)}"
                "[/blue] with [yellow]space[/yellow] installer was "
                "[green]successful[/green]"
            )
        else:
            rich.print(
                f"[red]Installer [yellow]{using}[/yellow] is not supported[/red]"
            )
            exit(-1)

def underscored(entry: str):
    return re.sub(r"[\W\s]", "_", entry)

def install_flow(
    reverse_id: str=typer.Argument(
        ...,
        show_default=False,
        help="Space Flow Reverse ID"
    ),
    deployment_id: str=typer.Option(
        None,
        '--deployment-id',
        '-d',
        show_default=False,
        help='Flow Deployment ID. If not set, will get active version flow.'
    ),
    branch: str=typer.Option(
        None,
        '--branch',
        '-b',
        show_default=False,
        help="Flow branch. If not specified, will take the active one."
    ),
    attach_any: bool=typer.Option(
        False,
        '--attach_any',
        '-a',
        show_default=False,
        help="Attach to any flow deployment"
    )
):
    installer = FlowInstaller()
    attach_to_last = deployment_id is None
    task = Space(
        reverse_id=reverse_id,
        attach_to_any=attach_any,
        attach_to_last=attach_to_last,
        deployment_id=deployment_id,
        branch=branch
    )
    version_name = task.component.version.readable_name
    mappings = {}
    injectables = task.get_injectables()
    for col in injectables:
        mappings[underscored(col.alias)] = col.alias

    integration = Integration(
        mapping=mappings,
        version=version_name,
        deployment=deployment_id,
        injectables=injectables
    )
    try:
        installer.install(
            reverse_id,
            integration
        )
    except Exception as e:
        rich.print(f"Failed to install [yellow]{reverse_id}[/yellow]: [red]{e}[/red]")
        exit(1)
    rich.print(
        f"Successfully installed: [yellow]{reverse_id}[/yellow], "
        f"version [blue]{version_name}[/blue]"
    )



import logging

import typer
import typer.core
from malevich_space.cli.cli import app as space_app

import malevich.help as help

from ._cli.ci import app as ci_app
from ._cli.core.app import core_app
from ._cli.dev.dev import dev as dev_app
from ._cli.flow import flow as flow_app
from ._cli.init import init as init_
from ._cli.install import auto_use
from ._cli.list import list_packages
from ._cli.manifest import app as manifest_app
from ._cli.new import new
from ._cli.prefs import prefs as prefs
from ._cli.remove import remove
from ._cli.restore import restore
from ._cli.space.init import init
from ._cli.space.login import login
from ._cli.space.upload import upload
from ._cli.space.whoami import get_user_on_space
from ._cli.use import use as use_app
from .constants import APP_HELP

logging.getLogger("gql.transport.requests").setLevel(logging.ERROR)
app = typer.Typer(
    help=APP_HELP,
    rich_markup_mode="rich",
)

# =============== Register Commands ===============
# -------------------------------------------------

# Malevich Commands
# -------------------------------------------------

groups = {
    'dm': 'Dependency Management',
    'pm': 'Project Management',
    'space': 'Space Management',
    'core': 'Engine (Core) API',
    'cfg': 'User Settings'
}

# malevich install
app.registered_commands.append(
    typer.models.CommandInfo(
        name="install",
        help=help.install["--help"],
        callback=auto_use,
        cls=typer.core.TyperCommand,
        rich_help_panel=groups['dm']
    )
)

# malevich restore
app.registered_commands.append(
    typer.models.CommandInfo(
        help=help.restore["--help"],
        callback=restore,
        cls=typer.core.TyperCommand,
        rich_help_panel=groups['dm']
    )
)

# malevich remove
app.registered_commands.append(
    typer.models.CommandInfo(
        help=help.remove["--help"],
        callback=remove,
        cls=typer.core.TyperCommand,
        rich_help_panel=groups['dm']
    )
)

# malevich list
app.registered_commands.append(
    typer.models.CommandInfo(
        name="list",
        help=help.list_["--help"],
        callback=list_packages,
        cls=typer.core.TyperCommand,
        rich_help_panel=groups['dm']
    )
)

# malevich new
app.registered_commands.append(
    typer.models.CommandInfo(
        name="new",
        help=help.new["--help"],
        callback=new,
        cls=typer.core.TyperCommand,
        rich_help_panel=groups['pm']
    )
)

# malevich init
app.registered_commands.append(
    typer.models.CommandInfo(
        name="init",
        help="Initializes new Malevich project",
        callback=init_,
        cls=typer.core.TyperCommand,
        rich_help_panel=groups['pm']
    )
)

# _________________________________________________

# Space Additional Commands
# -------------------------------------------------

# malevich space init
space_app.registered_commands.append(
    typer.models.CommandInfo(
        help=help.space["init --help"],
        callback=init,
        cls=typer.core.TyperCommand
    )
)

# malevich space login
space_app.registered_commands.append(
    typer.models.CommandInfo(
        help=help.space["login --help"],
        callback=login,
        cls=typer.core.TyperCommand
    )
)

space_app.registered_commands.append(
    typer.models.CommandInfo(
        "whoami",
        help=help.space["whoami --help"],
        callback=get_user_on_space,
        cls=typer.core.TyperCommand
    )
)


space_app.registered_commands.append(
    typer.models.CommandInfo(
        "upload-flow",
        help=help.space["upload-flow --help"],
        callback=upload,
        cls=typer.core.TyperCommand
    )
)
# __________________________________________________


# =============== Register Groups ==================
# --------------------------------------------------


# malevich space flow
space_app.add_typer(flow_app, name="flow")

# malevich use
app.add_typer(use_app, name="use", rich_help_panel=groups['dm'])

# malevich manifest
app.add_typer(manifest_app, name="manifest", rich_help_panel=groups["pm"])

# malevich space
app.add_typer(
    space_app,
    name="space",
    help=help.space["--help"],
    rich_help_panel=groups['space']
)

# malevich ci
app.add_typer(
    ci_app,
    name="ci",
    help=help.ci["--help"],
    rich_help_panel=groups['pm']
)

# malevich prefs
app.add_typer(prefs, name="prefs", rich_help_panel=groups['cfg'])

# malevich dev
app.add_typer(dev_app, name='dev', hidden=True)

#malevich core
app.add_typer(
    core_app,
    name='core',
    help=help.core['--help'],
    rich_help_panel=groups['core']
)
# _________________________________________________


class CLIContext:
    global_ = False

@app.callback()
def main_callback(
    ctx: typer.Context,
    global_: bool = typer.Option(
        False,
        "--global",
        "-g",
        help="Run the command in global context",
    ),
) -> None:
    ctx.obj = CLIContext()
    ctx.obj.global_ = global_

def main() -> None:
    """Entry point"""
    from malevich._cli.misc.manifest_to_env import manifest_as_env
    with manifest_as_env:
        app()


if __name__ == "__main__":
    main()



import os
import tempfile
from subprocess import run
from typing import Optional


def clone_python_files(
    link: str,
    auth: Optional[tuple[str, str]] = None,
    branch: Optional[str] = None,
    folder: os.PathLike = tempfile.TemporaryDirectory(),
) -> list[str]:
    """Clone a Git repository and return a list of Python files.

    The function is optimized to clone only the latest commit and only the
    files that are needed.


    Args:
        link: Link to Git repository or Docker image
        auth: Username and password or token for Git repository
        branch: Branch to clone. Defaults to the default branch.
        folder: Folder to clone into. Defaults to a temporary directory.

    Returns:
        - file_paths: list[str] - paths to Python files

    Raises:
        CalledProcessError: If the `git clone` command fails.
        AssertionError: If the link is not a Git repository
    """
    # Must be a Git repository, so
    # it must start with `git` or `http`
    assert any(
        [link.startswith("git"), link.startswith("http")]
    ), "Link must be a Git repository or Docker image"

    # Base command for cloning
    # --depth 1: Only clone the latest commit
    # --no-tags: Do not clone tags
    # --filter=blob:none: Do not clone blobs (large files)
    command = ["git", "clone", "--depth", "1",
               "--no-tags", "--filter=blob:none"]

    # Add branch if specified
    if branch:
        command += ["--branch", branch]

    # Add authentication if specified
    if auth and auth[0] and auth[1]:
        command += [f"https://{auth[0]}:{auth[1]}@{link}"]
    else:
        command += [link]

    # Add folder to clone into
    command += [os.path.abspath(
        os.path.join(
            os.getcwd(),
            folder,
        )
    )]

    # Run the command
    run(command, check=True)

    # Traverse the directory and subdirectories to find all Python files
    return [
        os.path.join(root, file)
        for root, dirs, files in os.walk(folder)
        for file in files
        if file.endswith(".py")
    ]


def clone(
    link: str,
    auth: Optional[tuple[str, str]] = None,
    branch: Optional[str] = None,
    folder: os.PathLike = tempfile.TemporaryDirectory().name,
) -> None:
    command = ["git", "clone",  "--filter=blob:none"]

    # Add authentication if specified
    if auth and auth[0] and auth[1]:
        command += [f"https://{auth[0]}:{auth[1]}@{link}"]
    else:
        command += [link]

    # Add branch if specified
    if branch:
        command += ["--branch", branch]

    # Add folder to clone into
    command += [os.path.abspath(
        os.path.join(
            os.getcwd(),
            folder,
        )
    )]

    # Run the command
    run(command, check=True)

    # Return list of all files
    return [
        os.path.join(root, file)
        for root, dirs, files in os.walk(folder)
        for file in files
    ]



from inspect import isclass
from typing import Type, TypeVar

from pydantic import BaseModel

from malevich._autoflow import traced
from malevich.models import DocumentNode, DocumentOverride

DocumentArgType = TypeVar('DocumentArgType', bound=BaseModel)
_TypeOrValue = Type[DocumentArgType] | DocumentArgType | Type[dict] | dict
class document:  # noqa: N801
    @staticmethod
    def override(
        data: DocumentArgType,
    ) -> DocumentOverride:
        return DocumentOverride(data=data)

    def __new__(
        cls,
        reverse_id: str,
        type_or_value: _TypeOrValue,
        *,
        alias: str | None = None,
    ) -> traced[DocumentNode]:
        if type_or_value is None:
            type_or_value = dict

        if isclass(type_or_value):
            if issubclass(type_or_value, BaseModel):
                return traced(
                    DocumentNode(
                        document_class=type_or_value,
                        reverse_id=reverse_id,
                        alias=alias
                    )
                )
            else:
                return traced(
                    DocumentNode(
                        document_class=dict,
                        reverse_id=reverse_id,
                        alias=alias
                    )
                )
        else:
            return traced(
                DocumentNode(
                    document=type_or_value,
                    document_class=type(type_or_value),
                    reverse_id=reverse_id,
                    alias=alias
                )
            )



main = {
    '--help': """Welcome to [bold magenta]Malevich Metascript[/bold magenta]! This is a CLI tool that
allows you to create and run multitask pipelines using [yellow]Malevich API[/yellow].
\n _____ \n
The tool includes[i] autoflow [/i]engine, that analyzes your code and turns it into a human-readable
pipeline;[i] installers [/i]that enable you to populate your flow with a huge variety of pre-built
components; and finally[i] interpreters [/i]that run your code on Malevich backend freeing you
from installing any dependency.

How you start:
\t1. Run  malevich use to install apps and get access to their functionality
\t2. Write a pipeline using simple Python scripts and @flow decorators
\t3. Connect to Malevich API and run your flow effortlessly

Check out [link=https://docs.malevich.ai/malevich]Malevich documentation[/link] for more details"""
}

use = {
    "--help": """Install apps to get access to the processors they provide""",
    'image --help': """Install apps using published Docker images.

    Usage examples:

    1. Install an app from a Dockerhub image user/malevich:app and expose it as [b]app[b] in the current environment

            malevich use image app user/malevich:app

    2. Install the same app, but from a privately hosted Malevich Core

            malevich use image app user/malevich:app --core-host https://core.malevich.ai --core-user admin --core-token kd923fn
    """,
    'space --help': """Install apps available at Malevich Space.

    Usage examples:

    1. Install [code i]my.openai.app[/code i] and expose it as [b]openai[b] in the current environment

            malevich use space openai my.openai.app
    """,
    'local --help': """Install apps from a local code.

    Usage examples:

    1. Install an app from a local code

            malevich use local app ./path/to/app
    """,
}

install =  {
    '--help': """The simplest way to install Malevich apps

    Usage examples:

        1. Install utility

            malevich install utility

        2. Install utility and openai

            malevich install utility openai

        3. Install utility and openai using the image installer

            malevich install utility openai --using image

        4. Install utility using the image installer and specify the image reference

            malevich install utility --using image --with-args image_ref=user/malevich:utility
    """
}


restore = {
    '--help': """Restore all installed apps manifested in the current environment"""
}

space = {
    '--help': """Communicate with Malevich Space - a public provider of Malevich Core""",
    'init --help': """Imports Malevich Space configuration into the current environment""",
    'login --help': """Interactive login to Malevich Space""",
    'whoami --help': """Get information about connected Space user""",
    'upload-flow --help': """Upload a new version of flow declared with @flow decorator""",
}

remove = {
    '--help': """Remove apps from the current environment""",
}

ci = {
    '--help': """Setup CI/CD pipeline for your repository""",
}

list_ = {
    '--help': """List all installed package stubs."""
}

new = {
    '--help': """Create a new app from a template""",
}

core = {
    '--help': """Manage your Core account"""
}



import re
import warnings
from typing import Any, Callable, Literal, ParamSpec, Type, overload

from malevich_space.ops import SpaceOps
from malevich_space.schema import SpaceSetup
from rich.prompt import Prompt

from malevich.core_api import check_auth
from malevich.interpreter.local import LocalInterpreter
from malevich.models.task.interpreted.local import LocalTask

from ._cli.space.login import login
from ._utility.space.auto_space_ops import get_auto_ops
from ._utility.space.get_core_creds import (
    get_core_creds_from_db,
    get_core_creds_from_setup,
)
from .constants import DEFAULT_CORE_HOST
from .interpreter.core import CoreInterpreter
from .interpreter.space import SpaceInterpreter
from .manifest import manf
from .models.exceptions import NoTaskToConnectError
from .models.flow_function import FlowFunction
from .models.task.interpreted.core import CoreTask
from .models.task.interpreted.space import SpaceTask
from .models.task.promised import PromisedTask
from .types import TracedNode, TracedNodes

FlowArgs = ParamSpec('FlowArgs')

class Core:
    """Deployment assistant that manages interactions with Malevich Core
     Check out the details: https://core.malevich.ai/
    """
    def __new__(
        cls: Type,
        task: PromisedTask | TracedNode | TracedNodes | FlowFunction = None,
        *task_args,
        pipeline_id: str | None = None,
        should_connect: bool = False,
        core_host: str | None = DEFAULT_CORE_HOST,
        user: str | None = None,
        access_key: str | None = None,
        **task_kwargs
    ) -> CoreTask:
        if not task and not pipeline_id:
            raise ValueError("No task or pipeline_id provided")

        if not user:
            try:
                user, access_key = get_core_creds_from_setup(
                    SpaceSetup(**manf.query('space', resolve_secrets=True))
                )
            except Exception:
                from malevich._cli.space.login import login
                if login():
                    user, access_key = get_core_creds_from_setup(
                        SpaceSetup(**manf.query('space', resolve_secrets=True))
                    )
                else:
                    try:
                        user, access_key = get_core_creds_from_db(user, core_host)
                    except Exception:
                        # TODO: Add Follow: {link}
                        raise ValueError("No user found. Use `Core(user=...)`")

        if not access_key:
            access_key = Prompt.ask(f"Password for {user} on Core:", password=True)

        try:
            check_auth(conn_url=core_host, auth=(user, access_key,))
        except Exception:
            raise

        if isinstance(task, FlowFunction):
            task: PromisedTask = task(*task_args, **task_kwargs)

        intepreter = CoreInterpreter(
            core_auth=(user, access_key), core_host=core_host
        )

        if pipeline_id:
            try:
                return intepreter.attach(
                    pipeline_id
                )
            except NoTaskToConnectError:
                if should_connect:
                    raise
                else:
                    return intepreter.attach(
                        pipeline_id,
                        only_fetch=True
                    )
        else:
            task.interpret(intepreter)
            return task.get_interpreted_task()


class Space:

    @staticmethod
    def fetch(reverse_id: str, ops: SpaceOps | None = None) -> tuple[dict, str, dict[str, str]]:  # noqa: E501
        if not ops:
            ops = get_auto_ops()

        info = ops.get_available_flows(reverse_id=reverse_id)
        flow_branch_version = {}
        active_branch = info['component']['activeBranch']['details']['name']
        active_versions = {}
        for branch_ in info['component']['branches']['edges']:
            branch_name = branch_['node']['details']['name']
            active_versions[branch_name] = branch_['node']['activeVersion']['flow']['details']['uid']  # noqa: E501
            flow_branch_version[branch_name] = {}
            for version_ in branch_['node']['versions']['edges']:
                version_name = version_['node']['details']['readableName']
                flow_branch_version[branch_name][version_name] = version_['node']['flow']['details']['uid']  # noqa: E501

        return flow_branch_version, active_branch, active_versions

    def __new__(
        cls,
        task: PromisedTask | TracedNode | TracedNodes | FlowFunction = None,
        *task_args,
        reverse_id: str | None = None,
        uid: str | None = None,
        deployment_id: str | None = None,
        branch: str | None = None,
        version: str | None = None,
        ops: SpaceOps | None = None,
        policy: Literal['no_use', 'only_use', 'use_or_new', 'fetch'] = 'use_or_new',
        **task_kwargs
    ) -> SpaceTask:

        setup = None
        if not ops:
            try:
                setup = SpaceSetup(**manf.query('space', resolve_secrets=True))
            except Exception:
                if not login():
                    raise Exception(
                        "Could not login you. Please, run `malevich space login` "
                        "manually and provide correct credentials."
                    )
                setup = SpaceSetup(**manf.query('space', resolve_secrets=True))
        else:
            setup = ops.space_setup

        ops = SpaceOps(space_setup=setup)

        if not uid and reverse_id is not None:
            flow_branch_version, active_branch, active_versions = cls.fetch(
                reverse_id=reverse_id,
                ops=ops
            )

            if branch is not None:
                if branch not in flow_branch_version:
                    raise ValueError(
                        f"Branch {branch} not found. Available branches: "
                        f" {list(flow_branch_version.keys())}")
            else:
                branch = active_branch

            if version is not None:
                if version not in flow_branch_version[branch]:
                    raise ValueError(
                        f"Version {version} not found in branch {branch}. "
                        f"Available versions for branch {branch}: "
                        f" {list(flow_branch_version[branch].keys())}"
                    )
                else:
                    uid = flow_branch_version[branch][version]
            else:
                uid = active_versions[branch]

        interpreter = SpaceInterpreter(
            setup=setup,
            ops=ops,
            # version_mode=version_mode
        )

        if isinstance(task, FlowFunction):
            task: PromisedTask = task(*task_args, **task_kwargs)

        if isinstance(task, PromisedTask):
            reverse_id = task._component.reverse_id

        if task is None:
            task = interpreter.attach(
                reverse_id=reverse_id,
                flow_uid=uid,
                deployment_id=deployment_id,
                search_deployments=(policy not in ['no_use', 'fetch']),
            )

            if policy == 'no_use':
                task.prepare()
                return task
            elif task.get_stage().value != 'started':
                if policy == 'only_use':
                    if deployment_id is not None:
                        raise Exception(
                            f"The deployment with ID {deployment_id} is not active "
                            "while policy was set to 'use_only'. "
                            "Provide active deployment, "
                            "or change policy to 'use_or_new' or 'no_use'"
                        )
                    else:
                        raise Exception(
                            "No active tasks found for 'use_only' policy. "
                            "You can change policy 'use_or_new' or 'no_use'."
                        )
                elif policy != 'fetch':
                    task.prepare()
            return task
        else:
            if branch is not None:
                warnings.warn("Space(branch=...) is not supported when @flow function is used.")  # noqa: E501
            if version is not None:
                warnings.warn("Space(version=...) is not supported when @flow function is used.")  # noqa: E501

            task.interpret(interpreter)
        return task.get_interpreted_task()


class Local:
    def __new__(
        cls,
        task: PromisedTask | TracedNode | TracedNodes | FlowFunction,
        *task_args,
        **task_kwargs
    ) -> LocalTask:
        interpreter = LocalInterpreter()
        if isinstance(task, FlowFunction):
            task: PromisedTask = task(*task_args, **task_kwargs)

        task.interpret(interpreter)
        return task.get_interpreted_task()


import importlib.util
import os
from pathlib import Path
from typing import Annotated

import rich
import rich.progress
import typer
import sys

from malevich._deploy import Space
from malevich.models.flow_function import FlowFunction

def import_file(filepath):
    module_name = os.path.splitext(os.path.basename(filepath))[0]
    
    spec = importlib.util.spec_from_file_location(module_name, filepath)
    if spec is None:
        return None
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)

    return module

def upload(
    source_file: Annotated[str, typer.Argument(..., help="Path to the source file"),],
    flow_reverse_id: Annotated[str, typer.Option(..., help="Flow reverse ID"),] = None,
    branch: Annotated[str, typer.Option(..., help="Branch name to upload to"),] = None,
) -> None:
    """Upload a flow to the Space"""

    # Check whether path is module definition or file
    module = import_file(source_file).__dict__
    if module is None:
        raise ValueError(f"Could not import {source_file}")
    # Importing the source file
    if flow_reverse_id:
        flows = [module[flow_reverse_id]]
    else:
        flows = [
            flow for name in module
            if isinstance(flow := module[name], FlowFunction)
        ]

    with rich.progress.Progress(rich.progress.SpinnerColumn()) as progress:
        task = progress.add_task("Uploading flows", total=len(flows))
        for flow in flows:
            if not isinstance(flow, FlowFunction):
                rich.print(f"Skipping {flow} as it is not a @flow function")
            else:
                Space(flow).upload(branch=branch)
            progress.update(task, advance=1)
            rich.print(f"Flow {flow.reverse_id} uploaded" + (" to branch " + branch if branch else ""))  # noqa: E501
    rich.print("Upload complete")




from malevich.core_api import FilesDirs, get_collection_objects


def get_asset_full_tree(
    base_path: str,
    conn_url: str,
    auth: tuple[str, str]
) -> FilesDirs:
    file_dirs = get_collection_objects(
        path=base_path,
        recursive=True,
        conn_url=conn_url,
        auth=auth,
    )

    return file_dirs



import logging
from enum import Enum

from rich.logging import RichHandler

from malevich.manifest import manf

from ..models.actions import Action
from ..models.preferences import LogFormat, UserPreferences, VerbosityLevel

prefs = manf.query('preferences')
prefs = UserPreferences(**prefs) if prefs else UserPreferences()
logger = logging.getLogger('malevich')

if not any([x.name == 'META_H' for x in logger.handlers]):
    for h in logger.handlers:
        logger.removeHandler(h)

    logger.setLevel(logging.DEBUG)
    handler = (
        logging.StreamHandler()
        if prefs.log_format == LogFormat.Plain
        else RichHandler(markup=True)
    )
    handler.setLevel(logging.DEBUG)
    handler.set_name("META_H")
    formatter = logging.Formatter("[b](%(name)s)[/b] %(message)s", datefmt="[%X]")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.propagate = False


class LogLevel(Enum):
    Info = "INFO"
    Warning = "WARNING"
    Error = "ERROR"
    Debug = "DEBUG"


def _cout(
    message: str,
    level: LogLevel = LogLevel.Info,
    *args
) -> None:
    if level == LogLevel.Info:
        logger.info(message, extra={"markup": True})
    elif level == LogLevel.Warning:
        logger.warning(message, extra={"markup": True})
    elif level == LogLevel.Error:
        logger.error(message, extra={"markup": True})
    elif level == LogLevel.Debug:
        logger.debug(message, extra={"markup": True})


def cout(
    action: Action,
    message: str,
    verbosity: VerbosityLevel = VerbosityLevel.Quiet,
    level: LogLevel = LogLevel.Info,
    *args,
) -> None:
    prefs: UserPreferences = manf.query('preferences')
    if not prefs:
        prefs = UserPreferences()
    else:
        prefs = UserPreferences(**prefs)
    if prefs.verbosity[action.value] == VerbosityLevel.Quiet:
        return
    elif verbosity.value <= prefs.verbosity[action.value]:
        if prefs.log_format == LogFormat.Plain:
            _cout(f"({action.name}) {message}", level=level, *args)
        else:
            _cout(f"[medium_purple2][{action.name}][/medium_purple2] {message}",
                  level=level,
                  *args
            )
    else:
        return



import malevich_coretools as api


class BaseCoreService:
    """Provides API access to the Malevich Core service."""

    def __init__(self, auth: api.AUTH, conn_url: str) -> None:
        """Provides API access to the Malevich Core service.

        The service captures the user's authentication credentials
        and the URL of the Malevich Core service and provides functions
        to interact with the service.

        Args:
            auth (malevich_coretools.AUTH): The user's authentication credentials.
            conn_url (str): The URL of the Malevich Core service.
        """
        self.auth = auth
        self.conn_url = conn_url
